"""
TODO: write doc
"""
__title__ = "Events"

from .badweather import BadWeatherEvent
from .exposurestatuschanged import ExposureStatusChangedEvent
from .event import Event
from .filterchanged import FilterChangedEvent
from .focusfound import FocusFoundEvent
from .goodweather import GoodWeatherEvent
from .log import LogEvent
from .moduleclosed import ModuleClosedEvent
from .moduleopened import ModuleOpenedEvent
from .motionstatuschanged import MotionStatusChangedEvent
from .move import MoveEvent, MoveRaDecEvent, MoveAltAzEvent
from .newimage import NewImageEvent
from .newspectrum import NewSpectrumEvent
from .roofclosing import RoofClosingEvent
from .roofopened import RoofOpenedEvent
from .taskstarted import TaskStartedEvent
from .taskfailed import TaskFailedEvent
from .taskfinished import TaskFinishedEvent
from .testevent import TestEvent
from .offsets import OffsetsEvent, OffsetsRaDecEvent, OffsetsAltAzEvent



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Tuple, List

from .interface import Interface


class IConfig(Interface, metaclass=ABCMeta):
    """The module allows access to some of its configuration options."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def get_config_caps(self, **kwargs: Any) -> Dict[str, Tuple[bool, bool, bool]]:
        """Returns dict of all config capabilities. First value is whether it has a getter, second is for the setter,
        third is for a list of possible options..

        Returns:
            Dict with config caps
        """
        ...

    @abstractmethod
    async def get_config_value(self, name: str, **kwargs: Any) -> Any:
        """Returns current value of config item with given name.

        Args:
            name: Name of config item.

        Returns:
            Current value.

        Raises:
            ValueError: If config item of given name does not exist.
        """
        ...

    @abstractmethod
    async def get_config_value_options(self, name: str, **kwargs: Any) -> List[str]:
        """Returns possible values for config item with given name.

        Args:
            name: Name of config item.

        Returns:
            Possible values.

        Raises:
            ValueError: If config item of given name does not exist.
        """
        ...

    @abstractmethod
    async def set_config_value(self, name: str, value: Any, **kwargs: Any) -> None:
        """Sets value of config item with given name.

        Args:
            name: Name of config item.
            value: New value.

        Raises:
            ValueError: If config item of given name does not exist or value is invalid.
        """
        ...


__all__ = ["IConfig"]



import io
import logging
import os
from typing import Optional, Dict, Any, Tuple, List, Callable, Type
import yaml
from astropy.io import fits
import pandas as pd

from pyobs.object import get_class_from_string
from pyobs.images import Image
from .file import VFSFile


log = logging.getLogger(__name__)


class VirtualFileSystem(object):
    """Base for a virtual file system."""

    __module__ = "pyobs.vfs"

    def __init__(self, roots: Optional[Dict[str, Any]] = None, **kwargs: Any):
        """Create a new VFS.

        Args:
            roots: Dictionary containing roots, see :mod:`~pyobs.vfs` for examples.
        """

        # if no root for 'pyobs' is given, add one
        self._roots: Dict[str, Any] = {
            "pyobs": {"class": "pyobs.vfs.LocalFile", "root": os.path.expanduser("~/.pyobs/")}
        }
        if roots is not None:
            self._roots.update(roots)

    @staticmethod
    def split_root(path: str) -> Tuple[str, str]:
        """Splits the root from the rest of the path.

        Args:
            path (str): Path to split.

        Returns:
            (tuple) Tuple (root, filename).
        """

        # remove leading slash
        if path.startswith("/"):
            path = path[1:]

        # no more slash left?
        if "/" not in path:
            raise ValueError("No valid path with a root.")

        # get position of first slash and split
        pos = path.index("/")
        root = path[:pos]
        filename = path[pos + 1 :]

        # return it
        return root, filename

    def open_file(self, filename: str, mode: str) -> VFSFile:
        """Open a file. The handling class is chosen depending on the rootse in the filename.

        Args:
            filename (str): Name of file to open.
            mode (str): Opening mode.

        Returns:
            (IOBase) File like object for given file.
        """
        # split root
        root, filename = VirtualFileSystem.split_root(filename)

        # does root exist?
        if root not in self._roots:
            raise ValueError("Could not find root {0} for file.".format(root))

        # create file object
        from pyobs.object import get_object

        fd = get_object(self._roots[root], object_class=VFSFile, name=filename, mode=mode)

        # return it
        return fd

    async def read_fits(self, filename: str) -> fits.HDUList:
        """Convenience function that wraps around open_file() to read a FITS file and put it into a astropy FITS
        structure.

        Args:
            filename: Name of file to download.

        Returns:
            A PrimaryHDU containing the FITS file.
        """
        async with self.open_file(filename, "rb") as f:
            data = await f.read()
            return fits.HDUList.fromstring(data)

    async def write_fits(self, filename: str, hdulist: fits.HDUList, *args: Any, **kwargs: Any) -> None:
        """Convenience function for writing an Image to a FITS file.

        Args:
            filename: Name of file to write.
            hdulist: hdu list to write.
        """

        # open file
        async with self.open_file(filename, "wb") as f:
            with io.BytesIO() as bio:
                hdulist.writeto(bio, *args, **kwargs)
                await f.write(bio.getvalue())

    async def read_image(self, filename: str) -> Image:
        """Convenience function that wraps around open_file() to read an Image.

        Args:
            filename: Name of file to download.

        Returns:
            An image object
        """
        async with self.open_file(filename, "rb") as f:
            data = await f.read()
            return Image.from_bytes(data)

    async def write_image(self, filename: str, image: Image, *args: Any, **kwargs: Any) -> None:
        """Convenience function for writing an Image to a FITS file.

        Args:
            filename: Name of file to write.
            image: Image to write.
        """

        # open file
        async with self.open_file(filename, "wb") as f:
            with io.BytesIO() as bio:
                image.writeto(bio, *args, **kwargs)
                await f.write(bio.getvalue())

    async def read_csv(self, filename: str, *args: Any, **kwargs: Any) -> pd.DataFrame:
        """Convenience function for reading a CSV file into a DataFrame.

        Args:
            filename: Name of file to read.

        Returns:
            DataFrame with content of file.
        """

        try:
            # open file
            async with self.open_file(filename, "r") as f:
                data = await f.read()
                return pd.read_csv(io.StringIO(data), *args, **kwargs)

        except pd.errors.EmptyDataError:
            # on error, return empty dataframe
            return pd.DataFrame()

    async def write_csv(self, filename: str, df: pd.DataFrame, *args: Any, **kwargs: Any) -> None:
        """Convenience function for writing a CSV file from a DataFrame.

        Args:
            filename: Name of file to write.
            df: DataFrame to write.
        """

        async with self.open_file(filename, "w") as f:
            # create a StringIO as temporary write target
            with io.StringIO() as sio:
                # write table to sio
                df.to_csv(sio, *args, **kwargs)

                # and write all content to file
                await f.write(sio.getvalue())

    async def read_yaml(self, filename: str) -> Any:
        """Convenience function for reading a YAML file into a dict.

        Args:
            filename: Name of file to read.

        Returns:
            Content of file.
        """

        # open file
        async with self.open_file(filename, "r") as f:
            # read YAML
            data = await f.read()
            return yaml.safe_load(io.StringIO(data))

    async def write_yaml(self, filename: str, data: Any) -> None:
        """Convenience function for writing a YAML file from a dict.

        Args:
            data: dict to write.
            filename: Name of file to write.
        """

        # open file
        async with self.open_file(filename, "w") as f:
            # create StringIO as temp storage
            with io.StringIO() as sio:
                # dump to StringIO
                yaml.dump(data, sio)

                # write file from StringIO
                await f.write(sio.getvalue())

    async def local_path(self, path: str) -> str:
        """Returns a local filename, but only, if path leads to a LocalFile.

        Args:
            path: Path to get local path for.

        Returns:
            Local path.

        Raises:
            ValueError if path does not lead to LocalFile.
        """
        from .localfile import LocalFile

        # get class
        klass, root, path = self._get_class(path)

        # local file?
        if not issubclass(klass, LocalFile):
            raise ValueError(f"Given path {path} is not a local path.")

        # get local path
        return await klass.local_path(path, **self._roots[root])

    def _get_class(self, path: str) -> Tuple[Type[VFSFile], str, str]:
        # split root
        root, path = VirtualFileSystem.split_root(path)

        # get root class
        return get_class_from_string(self._roots[root]["class"]), root, path

    def _get_method(self, path: str, method: str) -> Tuple[Callable[..., Any], str, str]:
        # split root
        klass, root, path = self._get_class(path)

        # get find method
        return getattr(klass, method), root, path

    async def find(self, path: str, pattern: str) -> List[str]:
        """Find a file in the given path.

        Args:
            path: Path to search in.
            pattern: Pattern to search for.

        Returns:
            List of found files.
        """

        # get method
        find, root, path = self._get_method(path, "find")

        # and call it
        return await find(path, pattern, **self._roots[root])

    async def listdir(self, path: str) -> List[str]:
        """Find a file in the given path.

        Args:
            path: Path to search in.
            pattern: Pattern to search for.

        Returns:
            List of found files.
        """

        # get method
        listdir, root, path = self._get_method(path, "listdir")

        # and call it
        return await listdir(path, **self._roots[root])

    async def exists(self, path: str) -> bool:
        """Checks, whether a given path or file exists.

        Args:
            path: Path to check.

        Returns:
            Whether it exists or not
        """

        # get method
        exists, root, path = self._get_method(path, "exists")

        # and call it
        return await exists(path, **self._roots[root])

    async def remove(self, path: str) -> bool:
        """Removes file with given path.

        Args:
            path: Path to delete.

        Returns:
            Success of deletion.
        """

        # get method
        remove, root, path = self._get_method(path, "remove")

        # and call it
        return await remove(path, **self._roots[root])


__all__ = ["VirtualFileSystem", "VFSFile"]



from typing import Any

from PyQt5 import QtWidgets, QtGui
import logging


class ModuleGui(QtWidgets.QMainWindow, logging.Handler):  # type: ignore
    def __init__(self, *args: Any, **kwargs: Any):
        QtWidgets.QMainWindow.__init__(self, *args, **kwargs)
        logging.Handler.__init__(self, *args, **kwargs)

        # basic stuff
        self.setWindowTitle("pytel")
        self.resize(900, 500)

        # main menu
        menu_main = self.menuBar()
        menu_file = menu_main.addMenu("&File")
        menu_file.addAction("E&xit", self.close)

        # add QPlainTextEdit
        self.log = QtWidgets.QTextBrowser(self)
        self.setCentralWidget(self.log)

        # font
        font = QtGui.QFont()
        font.setPointSize(12)
        self.log.setFont(font)

        # create log handler
        formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(filename)s:%(lineno)d %(message)s")
        logging.root.addHandler(self)
        self.setFormatter(formatter)

    def emit(self, record):
        # format entry
        log_entry = self.format(record)

        # colors?
        if "[ERROR]" in log_entry:
            log_entry = '<font color="red">%s</font>' % log_entry
        elif "[WARNING]" in log_entry:
            log_entry = '<font color="orange">%s</font>' % log_entry

        # append to log
        self.log.append(log_entry)

        # scroll to end
        self.log.verticalScrollBar().setValue(self.log.verticalScrollBar().maximum())



from typing import Optional, Any
from typing_extensions import TypedDict

from .event import Event


DataType = TypedDict("DataType", {"message": Optional[str]})


class TestEvent(Event):
    """Just a test event."""

    __module__ = "pyobs.events"

    def __init__(self, message: Optional[str] = None, **kwargs: Any):
        Event.__init__(self)
        self.data["message"] = message


__all__ = ["TestEvent"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface


class IRunning(Interface, metaclass=ABCMeta):
    """The module can be running."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def is_running(self, **kwargs: Any) -> bool:
        """Whether a service is running."""
        ...


__all__ = ["IRunning"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface


class ISyncTarget(Interface, metaclass=ABCMeta):
    """The module can synchronize a target, e.g. via a telescope control software behinde an
    :class:`~pyobs.interfaces.ITelescope`."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def sync_target(self, **kwargs: Any) -> None:
        """Synchronize device on current target."""
        ...


__all__ = ["ISyncTarget"]



"""
A module in *pyobs* is the smalles executable unit. The base class for all modules is
:class:`~pyobs.modules.Module`, so all modules should derive from this class and, usually, implement
one or more :class:`~pyobs.interfaces.Interface`\ s.

Modules are usually not created directly in code, but via a configuration file, which is a YAML file that directly
maps to the constructor of the module plus an additional entry ``class`` containing the full reference to the class
of the module to instantiate.

Take, for instance, the :class:`~pyobs.modules.test.StandAlone` class, which has two parameters in its
constructor: ``message`` and ``interval``. So a valid configuration file would look like this::

    class: pyobs.modules.test.Standalone
    message: Test
    interval: 5

Note that both parameters have default values, so both could be omitted from the configuration. So, this would also
work::

    class: pyobs.modules.test.Standalone
    message: Test

In this case, interval would have to its default value of 10.

.. note::
    Remember that the ``*args`` and ``**kwargs`` are always forwarded to the super class(es), so the constructor of
    a module *always* also provides the parameters from :class:`~pyobs.object.Object` and
    :class:`~pyobs.modules.Module`.

Quite often, a parameter will accept both an object of a given type and a dict. If this is the case, the dict must
be another class definition with a ``class`` keyword, referring to a class of the given type. See, for example, the
``comm`` parameter of :class:`~pyobs.modules.Module`: it takes both a dict and a :class:`pyobs.comm.Comm`.
So in a configuration file, we can always specify a Comm object like this::

    comm:
        class: pyobs.comm.sleekxmpp.XmppComm
        jid: someone@example.com
        password: secret

An object of type :class:`~pyobs.comm.sleekxmpp.XmppComm` (which is a class derived from
:class:`~pyobs.comm.Comm`) will automatically be created.

With a Comm object, proxies to other modules can easily be created (see :mod:`~pyobs.comm` for details)::

    camera: ICamera = self.comm['camera']
    camera.expose()

Sometimes, multiple modules have to run in a single process, so that they can access a common resource. For this case
a :class:`~pyobs.modules.MultiModule` can contain multiple module descriptions.
"""

from .module import Module, MultiModule, timeout

__all__ = ["Module", "MultiModule", "timeout"]



from abc import ABCMeta, abstractmethod
from typing import Tuple, Any, Dict

from .IStartStop import IStartStop
from pyobs.utils.enums import WeatherSensors


class IWeather(IStartStop, metaclass=ABCMeta):
    """The module acts as a weather station."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def get_weather_status(self, **kwargs: Any) -> Dict[str, Any]:
        """Returns status of object in form of a dictionary. See other interfaces for details."""
        ...

    @abstractmethod
    async def is_weather_good(self, **kwargs: Any) -> bool:
        """Whether the weather is good to observe."""
        ...

    @abstractmethod
    async def get_current_weather(self, **kwargs: Any) -> Dict[str, Any]:
        """Returns current weather.

        Returns:
            Dictionary containing entries for time, good, and sensor, with the latter being another dictionary
            with sensor information, which contain a value and a good flag.
        """
        ...

    @abstractmethod
    async def get_sensor_value(self, station: str, sensor: WeatherSensors, **kwargs: Any) -> Tuple[str, float]:
        """Return value for given sensor.

        Args:
            station: Name of weather station to get value from.
            sensor: Name of sensor to get value from.

        Returns:
            Tuple of current value of given sensor or None and time of measurement or None.
        """
        ...


__all__ = ["IWeather"]



from .pyobs import init_cli, parse_cli, run


def main():
    from pyobs.application import GuiApplication

    # init argument parsing and parse it
    parser = init_cli()
    args = parse_cli(parser)

    # run app
    run(app_class=GuiApplication, **args)


if __name__ == "__main__":
    main()



from typing import Optional, Any
from typing_extensions import TypedDict

from pyobs.events.event import Event


DataType = TypedDict(
    "DataType", {"time": str, "level": str, "filename": str, "function": str, "line": int, "message": str}
)


class LogEvent(Event):
    """Event for log entries."""

    __module__ = "pyobs.events"

    def __init__(self, time: str, level: str, filename: str, function: str, line: int, message: str, **kwargs: Any):
        Event.__init__(self)
        self.data: DataType = {
            "time": time,
            "level": level,
            "filename": filename,
            "function": function,
            "line": line,
            "message": message,
        }

    @property
    def time(self) -> str:
        return str(self.data["time"])

    @property
    def level(self) -> str:
        return str(self.data["level"])

    @property
    def filename(self) -> str:
        return str(self.data["filename"])

    @property
    def function(self) -> str:
        return str(self.data["function"])

    @property
    def line(self) -> Optional[int]:
        return int(self.data["line"])

    @property
    def message(self) -> str:
        return str(self.data["message"])


__all__ = ["LogEvent"]



from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import Any, Optional

from .IReady import IReady
from pyobs.utils.enums import MotionStatus


class IMotion(IReady, metaclass=ABCMeta):
    """The module controls a device that can move."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def init(self, **kwargs: Any) -> None:
        """Initialize device.

        Raises:
            InitError: If device could not be initialized.
        """
        ...

    @abstractmethod
    async def park(self, **kwargs: Any) -> None:
        """Park device.

        Raises:
            ParkError: If device could not be parked.
        """
        ...

    @abstractmethod
    async def get_motion_status(self, device: Optional[str] = None, **kwargs: Any) -> MotionStatus:
        """Returns current motion status.

        Args:
            device: Name of device to get status for, or None.

        Returns:
            A string from the Status enumerator.
        """
        ...

    @abstractmethod
    async def stop_motion(self, device: Optional[str] = None, **kwargs: Any) -> None:
        """Stop the motion.

        Args:
            device: Name of device to stop, or None for all.
        """
        ...


__all__ = ["IMotion"]



class PID:
    """
    Discrete PID control
    """

    def __init__(
        self,
        p: float = 2.0,
        i: float = 0.0,
        d: float = 1.0,
        derivator: float = 0,
        integrator: float = 0,
        integrator_max: float = 500.0,
        integrator_min: float = -500.0,
    ):
        # public variables
        self.Kp = p
        self.Ki = i
        self.Kd = d
        self.Derivator = derivator
        self.Integrator = integrator
        self.Integrator_max = integrator_max
        self.Integrator_min = integrator_min

        # read-only
        self._set_point = 0.0
        self._error = 0.0

    def update(self, current_value: float) -> float:
        """Calculate PID output value for given reference input and feedback

        Args:
            current_value: last measured value.
        """

        self._error = current_value - self._set_point
        P_value = self.Kp * self.error
        D_value = self.Kd * (self.error - self.Derivator)
        self.Derivator = self.error
        self.Integrator += self.error
        if self.Integrator > self.Integrator_max:
            self.Integrator = self.Integrator_max
        elif self.Integrator < self.Integrator_min:
            self.Integrator = self.Integrator_min
        I_value = self.Integrator * self.Ki
        return P_value + I_value + D_value

    @property
    def setpoint(self) -> float:
        return self._set_point

    @setpoint.setter
    def setpoint(self, v: float) -> None:
        self._set_point = v
        self.Integrator = 0
        self.Derivator = 0

    @property
    def error(self) -> float:
        return self._error


__all__ = ["PID"]



import asyncio
import logging
from typing import Any

from pyobs.modules import Module


log = logging.getLogger(__name__)


class StandAlone(Module):
    """Example module that only logs the given message forever in the given interval."""

    __module__ = "pyobs.modules.test"

    def __init__(self, message: str = "Hello world", interval: int = 10, **kwargs: Any):
        """Creates a new StandAlone object.

        Args:
            message: Message to log in the given interval.
            interval: Interval between messages.
        """
        Module.__init__(self, **kwargs)

        # add thread func
        self.add_background_task(self._message_func)

        # store
        self._message = message
        self._interval = interval

    async def _message_func(self):
        """Thread function for async processing."""
        # loop until closing
        while True:
            # log message
            log.info(self._message)

            # sleep a little
            await asyncio.sleep(self._interval)


__all__ = ["StandAlone"]



from abc import ABCMeta, abstractmethod


class Interface(object, metaclass=ABCMeta):
    """Base class for all interfaces in pyobs."""

    __module__ = "pyobs.interfaces"
    pass


__all__ = ["Interface"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface


class IScriptRunner(Interface, metaclass=ABCMeta):
    """The module can execute a script."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def run_script(self, script: str, **kwargs: Any) -> None:
        """Run the given script.

        Args:
            script: Script to run.
        """
        ...


__all__ = ["IScriptRunner"]



import logging
from typing import Any

from pyobs.images.processor import ImageProcessor
from pyobs.images import Image


log = logging.getLogger(__name__)


class SoftBin(ImageProcessor):
    """Bin an image."""

    __module__ = "pyobs.images.processors.misc"

    def __init__(self, binning: int = 2, **kwargs: Any):
        """Init a new software binning pipeline step.

        Args:
            binning: Binning to apply to image.
        """
        ImageProcessor.__init__(self, **kwargs)

        # store
        self.binning = binning

    async def __call__(self, image: Image) -> Image:
        """Bin an image.

        Args:
            image: Image to bin.

        Returns:
            Binned image.
        """

        # copy image
        img = image.copy()
        if img.data is None:
            log.warning("No data found in image.")
            return image

        # calculate new shape, in which all binned pixels are in a higher dimension
        shape = (img.data.shape[0] // self.binning, self.binning, img.data.shape[1] // self.binning, self.binning)

        # reshape and average
        img.data = img.data.reshape(shape).mean(-1).mean(1)
        if img.data is None:
            log.warning("No data found in image after reshaping.")
            return image

        # set NAXIS1/2
        img.header["NAXIS2"], img.header["NAXIS1"] = img.data.shape

        # divide some header entries by binning
        for key in ["CRPIX1", "CRPIX2"]:
            if key in img.header:
                img.header[key] /= self.binning

        # multiply some header entries with binning
        for key in ["DET-BIN1", "DET-BIN2", "XBINNING", "YBINNING", "CDELT1", "CDELT2"]:
            if key in img.header:
                img.header[key] *= self.binning

        # return result
        return img


__all__ = ["SoftBin"]



import logging
from typing import Any
from astropy.coordinates import SkyCoord
import astropy.units as u

from pyobs.interfaces import ITelescope
from pyobs.utils.time import Time
from .base import SkyFlatsBasePointing
from ...parallel import Future

log = logging.getLogger(__name__)


class SkyFlatsStaticPointing(SkyFlatsBasePointing):
    """Static flat pointing."""

    __module__ = "pyobs.utils.skyflats.pointing"

    def __init__(self, initialized: bool = False, *args: Any, **kwargs: Any):
        """Inits new static pointing for sky flats.

        Args:
            initialized: If False, telescope does not move at all.
        """

        SkyFlatsBasePointing.__init__(self, *args, **kwargs)

        # whether we've moved already
        self._initialized = initialized

    async def __call__(self, telescope: ITelescope) -> None:
        """Move telescope.

        Args:
            telescope: Telescope to use.

        Returns:
            Future for the movement call.
        """

        if self._initialized:
            return
        self._initialized = True

        # calculate Alt/Az position of sun
        now = Time.now()
        sun = self.observer.sun_altaz(now)
        log.info("Sun is currently located at alt=%.2f°, az=%.2f°", sun.alt.degree, sun.az.degree)

        # get sweet spot for flat-fielding
        altaz = SkyCoord(
            alt=80 * u.deg, az=sun.az + 180 * u.degree, obstime=now, location=self.observer.location, frame="altaz"
        )
        log.info("Sweet spot for flat fielding is at alt=80°, az=%.2f°", altaz.az.degree)

        # move telescope
        log.info("Moving telescope to Alt=80, Az=%.2f...", altaz.az.degree)
        return await telescope.move_altaz(80, float(altaz.az.degree))

    async def reset(self) -> None:
        """Reset pointing."""
        self._initialized = False


__all__ = ["SkyFlatsStaticPointing"]



from __future__ import annotations
import logging
import math
import os
from collections import Coroutine
from typing import Union, Dict, Any, Tuple, Optional, List, cast
import astropy.units as u
from astropy.io import fits

from pyobs.images import Image
from pyobs.interfaces import IFitsHeaderBefore, IFitsHeaderAfter
from pyobs.modules import Module
from pyobs.utils.fits import format_filename
from pyobs.utils import exceptions as exc
from pyobs.utils.time import Time

log = logging.getLogger(__name__)


class FitsHeaderMixin:
    """Helper methods for all modules that implement IImageGrabber."""

    __module__ = "pyobs.mixins"

    def __init__(
        self,
        fits_namespaces: Optional[List[str]] = None,
        fits_headers: Optional[Dict[str, Any]] = None,
        filenames: str = "/cache/pyobs-{DAY-OBS|date:}-{FRAMENUM|string:04d}.fits",
        frame_number: bool = True,
        night_obs: bool = True,
        **kwargs: Any,
    ):
        """Initialise the mixin.

        Args:
            fits_namespaces: List of namespaces for FITS headers that this camera should request.
            fits_headers: Additional FITS headers.
            filename: Filename pattern for FITS images.
            frame_number: Whether to add frame number to FITS file header.
            night_obs: If True, DAY-OBS will contain the night of observation, not the calendar day.
        """
        module = cast(Module, self)

        # store
        self._fitsheadermixin_fits_namespaces = fits_namespaces
        self._fitsheadermixin_filename_pattern = filenames
        self._fitsheadermixin_fits_headers = fits_headers if fits_headers is not None else {}
        if "OBSERVER" not in self._fitsheadermixin_fits_headers:
            self._fitsheadermixin_fits_headers["OBSERVER"] = ["pyobs", "Name of observer"]
        self._fitsheadermixin_night_obs = night_obs

        # night exposure number
        self._fitsheadermixin_cache = "/pyobs/modules/%s/cache.yaml" % module.name()
        self._fitsheadermixin_enable_frame_number = frame_number
        self._fitsheadermixin_frame_number = 0

    async def request_fits_headers(self, before: bool = True) -> Dict[str, Coroutine]:
        """Request FITS headers from other modules.

        Returns:
            Futures from all modules.
        """

        # init
        futures: Dict[str, Coroutine] = {}

        # we can only do this with a comm module
        module = cast(Module, self)
        if module.comm:
            # get clients that provide fits headers
            clients = await module.comm.clients_with_interface(IFitsHeaderBefore if before else IFitsHeaderAfter)

            # create and run a threads in which the fits headers are fetched
            for client in clients:
                log.debug("Requesting FITS headers from %s...", client)
                if before:
                    proxy1 = await module.proxy(client, IFitsHeaderBefore)
                    futures[client] = proxy1.get_fits_header_before(self._fitsheadermixin_fits_namespaces)
                else:
                    proxy2 = await module.proxy(client, IFitsHeaderAfter)
                    futures[client] = proxy2.get_fits_header_after(self._fitsheadermixin_fits_namespaces)

        # finished
        return futures

    async def add_requested_fits_headers(
        self, image: Union[Image, fits.PrimaryHDU], futures: Dict[str, Coroutine]
    ) -> None:
        """Add requested FITS headers to header of given image.

        Args:
            image: Image with header to add to.
            futures: Futures to get headers from.
        """

        # get fits headers from other clients
        for client, future in futures.items():
            # join thread
            log.info("Fetching FITS headers from %s...", client)
            try:
                headers = await future
            except exc.RemoteError as e:
                log.warning("Could not fetch FITS headers from %s: %s.", client, str(e))
                continue

            # add them to fits file
            if headers:
                log.debug("Adding additional FITS headers from %s..." % client)
                for key, value in headers.items():
                    # if value is not a string, it may be a list of value and comment
                    if isinstance(value, list) and not isinstance(value, str):
                        # convert list to tuple
                        image.header[key] = tuple(value)
                    else:
                        image.header[key] = value

    async def add_fits_headers(self, image: Union[Image, fits.PrimaryHDU]) -> None:
        """Add requested FITS headers to header of given image.

        Args:
            image: Image with header to add to.
        """

        # add HDU name
        image.header["EXTNAME"] = "SCI"

        # add static fits headers
        for key, value in self._fitsheadermixin_fits_headers.items():
            image.header[key] = tuple(value)

        # add more fits headers
        self._fitsheadermixin_add_fits_headers(image)
        if self._fitsheadermixin_enable_frame_number:
            await self._fitsheadermixin_add_framenum(image)

    def _fitsheadermixin_add_fits_headers(self, image: Union[Image, fits.PrimaryHDU]) -> None:
        """Add FITS header keywords to the given FITS header.

        Args:
            image: Image with header to add to.
        """
        module = cast(Module, self)

        # get header
        hdr = image.header

        # we definitely need a DATE-OBS and IMAGETYP!!
        if "DATE-OBS" not in hdr:
            log.warning("No DATE-OBS found in FITS header, adding NO further information!")
            return

        # get date obs
        date_obs = Time(hdr["DATE-OBS"])

        # UT1-UTC
        hdr["UT1_UTC"] = (float(date_obs.delta_ut1_utc), "UT1-UTC")

        # basic stuff
        hdr["EQUINOX"] = (2000.0, "Equinox of celestial coordinate system")

        # do we have a location?
        if module.location is not None:
            loc = module.location
            # add location of telescope
            hdr["LONGITUD"] = (float(loc.lon.degree), "Longitude of the telescope [deg E]")
            hdr["LATITUDE"] = (float(loc.lat.degree), "Latitude of the telescope [deg N]")
            hdr["HEIGHT"] = (float(loc.height.value), "Altitude of the telescope [m]")

            # add local sidereal time
            if module.observer is not None:
                lst = module.observer.local_sidereal_time(date_obs)
                hdr["LST"] = (lst.to_string(unit=u.hour, sep=":"), "Local sidereal time")

        # date of night this observation is in
        if self._fitsheadermixin_night_obs:
            hdr["DAY-OBS"] = (date_obs.night_obs(module.observer).strftime("%Y-%m-%d"), "Night of observation")
        else:
            hdr["DAY-OBS"] = (date_obs.strftime("%Y-%m-%d"), "Day of observation")

    async def _fitsheadermixin_add_framenum(self, image: Union[Image, fits.PrimaryHDU]) -> None:
        """Add FRAMENUM keyword to header

        Args:
            image: Image with header to add to.
        """
        module = cast(Module, self)

        # get header
        hdr = image.header

        # get night from header
        night = hdr["DAY-OBS"]

        # increase night exp
        self._fitsheadermixin_frame_number += 1

        # do we have a cache?
        if self._fitsheadermixin_cache is not None:
            # try to load it
            try:
                # load cache
                cache = await module.vfs.read_yaml(self._fitsheadermixin_cache)

                # get new number
                if cache is not None and "framenum" in cache:
                    self._fitsheadermixin_frame_number = cache["framenum"] + 1

                # if nights differ, reset count
                if cache is not None and "night" in cache and night != cache["night"]:
                    self._fitsheadermixin_frame_number = 1

            except (FileNotFoundError, ValueError):
                pass

            # write file
            try:
                await module.vfs.write_yaml(
                    self._fitsheadermixin_cache, {"night": night, "framenum": self._fitsheadermixin_frame_number}
                )
            except (FileNotFoundError, ValueError):
                log.warning("Could not write camera cache file.")

        # set it
        hdr["FRAMENUM"] = self._fitsheadermixin_frame_number

    def format_filename(self, image: Union[Image, fits.PrimaryHDU]) -> Optional[str]:
        """Format filename according to given pattern and store in header of image.

        Args:
            image: Image with header to add to.
        """

        # no pattern?
        if self._fitsheadermixin_filename_pattern is None:
            return None

        # create a temporary filename
        filename = format_filename(image.header, self._fitsheadermixin_filename_pattern)
        if filename is None:
            return None

        # set it
        image.header["ORIGNAME"] = (os.path.basename(filename), "The original file name")
        image.header["FNAME"] = (os.path.basename(filename), "FITS file file name")
        return filename


class ImageFitsHeaderMixin(FitsHeaderMixin):
    """Helper methods for all modules that need FITS headers for an image."""

    __module__ = "pyobs.mixins"

    def __init__(self, centre: Optional[Tuple[float, float]] = None, rotation: float = 0.0, **kwargs: Any):
        """Initialise the mixin.

        Args:
            fits_namespaces: List of namespaces for FITS headers that this camera should request.
            fits_headers: Additional FITS headers.
            centre: (x, y) tuple of camera centre.
            rotation: Rotation east of north.
            filename: Filename pattern for FITS images.
        """
        FitsHeaderMixin.__init__(self, **kwargs)

        # store
        self._fitsheadermixin_centre = centre
        self._fitsheadermixin_rotation = rotation

    @property
    def rotation(self) -> float:
        return self._fitsheadermixin_rotation

    @property
    def centre(self) -> Optional[Tuple[float, float]]:
        return self._fitsheadermixin_centre

    def _fitsheadermixin_add_fits_headers(self, image: Union[Image, fits.PrimaryHDU]) -> None:
        """Add FITS header keywords to the given FITS header.

        Args:
            image: Image with header to add to.
        """

        # parent method
        FitsHeaderMixin._fitsheadermixin_add_fits_headers(self, image)

        # get header
        hdr = image.header

        # convenience function to return value of keyword
        def v(k: str) -> Any:
            return hdr[k][0] if isinstance(k, list) or isinstance(k, tuple) else hdr[k]

        # set CRVAL1/2 from RA/Dec
        if "TEL-RA" in hdr and "TEL-DEC" in hdr:
            hdr["CRVAL1"] = hdr["TEL-RA"]
            hdr["CRVAL2"] = hdr["TEL-DEC"]

        # pixel size in world coordinates
        if "DET-PIXL" in hdr and "TEL-FOCL" in hdr and "DET-BIN1" in hdr and "DET-BIN2" in hdr:
            tmp = 360.0 / (2.0 * math.pi) * v("DET-PIXL") / v("TEL-FOCL")
            hdr["CDELT1"] = (-tmp * v("DET-BIN1"), "Coordinate increment on x-axis [deg/px]")
            hdr["CDELT2"] = (+tmp * v("DET-BIN2"), "Coordinate increment on y-axis [deg/px]")
            hdr["CUNIT1"] = ("deg", "Units of CRVAL1, CDELT1")
            hdr["CUNIT2"] = ("deg", "Units of CRVAL2, CDELT2")
            hdr["WCSAXES"] = (2, "Number of WCS axes")
        else:
            log.warning("Could not calculate CDELT1/CDELT2 (DET-PIXL/TEL-FOCL/DET-BIN1/DET-BIN2 missing).")

        # centre pixel
        if self._fitsheadermixin_centre is not None:
            hdr["DET-CPX1"] = (self._fitsheadermixin_centre[0], "x-pixel on mechanical axis in unbinned image")
            hdr["DET-CPX2"] = (self._fitsheadermixin_centre[1], "y-pixel on mechanical axis in unbinned image")
        else:
            log.warning("Could not calculate DET-CPX1/DET-CPX2 (centre not given in config).")

        # reference pixel in binned image
        if "DET-CPX1" in hdr and "DET-BIN1" in hdr and "DET-CPX2" in hdr and "DET-BIN2" in hdr:
            # offset?
            off_x, off_y = 0, 0
            if "XORGSUBF" in hdr and "YORGSUBF" in hdr:
                off_x = v("XORGSUBF") if "XORGSUBF" in hdr else 0.0
                off_y = v("YORGSUBF") if "YORGSUBF" in hdr else 0.0
            hdr["CRPIX1"] = ((v("DET-CPX1") - off_x) / v("DET-BIN1"), "Reference x-pixel position in binned image")
            hdr["CRPIX2"] = ((v("DET-CPX2") - off_y) / v("DET-BIN2"), "Reference y-pixel position in binned image")
        else:
            log.warning(
                "Could not calculate CRPIX1/CRPIX2 " "(XORGSUBF/YORGSUBF/DET-CPX1/TEL-CPX2/DET-BIN1/DET-BIN2) missing."
            )
        # only add all this stuff for OBJECT images
        if "IMAGETYP" not in hdr or hdr["IMAGETYP"] not in ["dark", "bias"]:
            # projection
            hdr["CTYPE1"] = ("RA---TAN", "RA in tangent plane projection")
            hdr["CTYPE2"] = ("DEC--TAN", "Dec in tangent plane projection")

            # PC matrix: rotation only, shift comes from CDELT1/2
            if self._fitsheadermixin_rotation is not None:
                theta_rad = math.radians(self._fitsheadermixin_rotation)
                cos_theta = math.cos(theta_rad)
                sin_theta = math.sin(theta_rad)
                hdr["PC1_1"] = (+cos_theta, "Partial of first axis coordinate w.r.t. x")
                hdr["PC1_2"] = (-sin_theta, "Partial of first axis coordinate w.r.t. y")
                hdr["PC2_1"] = (+sin_theta, "Partial of second axis coordinate w.r.t. x")
                hdr["PC2_2"] = (+cos_theta, "Partial of second axis coordinate w.r.t. y")
            else:
                log.warning("Could not calculate CD matrix (rotation or CDELT1/CDELT2 missing.")


class SpectrumFitsHeaderMixin(FitsHeaderMixin):
    """Helper methods for all modules that need FITS headers for an image."""

    __module__ = "pyobs.mixins"
    pass


__all__ = ["FitsHeaderMixin", "ImageFitsHeaderMixin", "SpectrumFitsHeaderMixin"]



import logging
import os
from typing import TYPE_CHECKING, Any

from pyobs.events import LogEvent

if TYPE_CHECKING:
    from pyobs.comm import Comm


class CommLoggingHandler(logging.Handler):
    """A logging handler that sends all messages through a Comm module."""

    def __init__(self, comm: "Comm"):
        """Create a new logging handler.

        Args:
            comm: Comm module to use.
        """
        logging.Handler.__init__(self)
        self._comm = comm

        # get formatter
        self._formatter = logging.Formatter()

    def emit(self, rec: Any) -> None:
        """Send a new log entry to the comm module.

        Args:
            rec: Log record to send.
        """

        # format message
        msg = self._formatter.format(rec)

        # create and send event
        entry = LogEvent(rec.created, rec.levelname, os.path.basename(rec.pathname), rec.funcName, rec.lineno, msg)
        self._comm.log_message(entry)


__all__ = ["CommLoggingHandler"]



from sleekxmpp.plugins.xep_0009 import XEP_0009 as XEP_0009_original
from sleekxmpp.xmlstream import ET


class XEP_0009(XEP_0009_original):
    """Small fix for the original XEP_0009 plugin."""

    def _handle_error(self, iq):
        pass

    def extract_method(self, stanza):
        xml = ET.fromstring("%s" % stanza)
        return xml.find("./{jabber:iq:rpc}methodCall/{jabber:iq:rpc}methodName").text

    def item_not_found(self, iq):
        """Expose method to public."""
        return self._item_not_found(iq)

    def send_fault(self, iq, fault_xml):
        """Expose method to public."""
        return self._send_fault(iq, fault_xml)



import logging
import asyncio
from typing import Union, List, Any, Optional

from pyobs.modules import Module
from pyobs.events import NewImageEvent, Event
from pyobs.utils.fits import format_filename

log = logging.getLogger(__name__)


class ImageWriter(Module):
    """Writes new images to disk."""

    __module__ = "pyobs.modules.image"

    def __init__(
        self, filename: str = "/archive/{FNAME}", sources: Optional[Union[str, List[str]]] = None, **kwargs: Any
    ):
        """Creates a new image writer.

        Args:
            filename: Pattern for filename to store images at.
            sources: List of sources (e.g. cameras) to process images from or None for all.
        """
        Module.__init__(self, **kwargs)

        # add thread func
        self.add_background_task(self._worker, True)

        # variables
        self._filename = filename
        self._sources = [sources] if isinstance(sources, str) else sources
        self._queue = asyncio.Queue()

    async def open(self) -> None:
        """Open image writer."""
        await Module.open(self)

        # subscribe to channel with new images
        if self.comm is not None:
            log.info("Subscribing to new image events...")
            await self.comm.register_event(NewImageEvent, self.process_new_image_event)

    async def process_new_image_event(self, event: Event, sender: str) -> bool:
        """Puts a new images in the DB with the given ID.

        Args:
            event:  New image event
            sender: Who sent the event?

        Returns:
            Success
        """
        if not isinstance(event, NewImageEvent):
            return False

        # filter by source
        if self._sources is not None and sender not in self._sources:
            return False

        # queue file
        log.info("Received new image event from %s.", sender)
        self._queue.put_nowait(event.filename)
        return True

    async def _worker(self) -> None:
        """Worker thread."""

        # run forever
        while True:
            # get next filename
            filename = await self._queue.get()

            try:
                # download image
                log.info("Downloading file %s...", filename)
                img = await self.vfs.read_image(filename)
            except FileNotFoundError:
                log.error("Could not download image.")
                continue

            # output filename
            try:
                output = format_filename(img.header, self._filename)
            except KeyError as e:
                log.error("Could not format filename: %s", e)
                continue

            try:
                # open output
                log.info("Storing image as %s...", output)
                await self.vfs.write_image(output, img)
            except Exception:
                log.error("Could not store image.")


__all__ = ["ImageWriter"]



import logging
from typing import Dict, Any, Optional

from pyobs.robotic.scripts import Script
from pyobs.robotic import TaskSchedule, TaskArchive, TaskRunner

log = logging.getLogger(__name__)


class LcoScript(Script):
    """Auto SCRIPT script for LCO configs."""

    def __init__(self, scripts: Dict[str, Script], **kwargs: Any):
        """Initialize a new LCO auto focus script.

        Args:
            scripts: External scripts to run
        """
        Script.__init__(self, **kwargs)

        # store
        self.scripts = scripts

    def _get_config_script(self, config: Dict[str, Any]) -> Script:
        """Get config script for given configuration.

        Args:
            config: Config to create runner for.

        Returns:
            Script for running config

        Raises:
            ValueError: If could not create runner.
        """

        # what do we run?
        config_type = config["extra_params"]["script_name"]
        if config_type not in self.scripts:
            raise ValueError('No script found for script type "%s".' % config_type)

        # create script handler
        return self.get_object(self.scripts[config_type], Script, configuration=config)

    async def can_run(self) -> bool:
        """Checks, whether this task could run now.

        Returns:
            True, if task can run now.
        """

        # get config runner
        runner = self._get_config_script(self.configuration)

        # if any runner can run, we proceed
        return await runner.can_run()

    async def run(
        self,
        task_runner: TaskRunner,
        task_schedule: Optional[TaskSchedule] = None,
        task_archive: Optional[TaskArchive] = None,
    ) -> None:
        """Run script.

        Raises:
            InterruptedError: If interrupted
        """

        # get config runner
        runner = self._get_config_script(self.configuration)

        # run it
        await runner.run(task_runner=task_runner, task_schedule=task_schedule, task_archive=task_archive)


__all__ = ["LcoScript"]



from typing_extensions import TypedDict
from typing import Any

from pyobs.events import Event


class MoveEvent(Event):
    """Event to be sent when moving to a new target."""

    __module__ = "pyobs.events"
    pass


DataTypeRaDec = TypedDict("DataTypeRaDec", {"ra": float, "dec": float})


class MoveRaDecEvent(MoveEvent):
    """Event to be sent when moving to RA/Dec."""

    __module__ = "pyobs.events"

    def __init__(self, ra: float, dec: float, **kwargs: Any):
        MoveEvent.__init__(self)
        self.data: DataTypeRaDec = {"ra": ra, "dec": dec}

    @property
    def ra(self) -> float:
        return self.data["ra"]

    @property
    def dec(self) -> float:
        return self.data["dec"]


DataTypeAltAz = TypedDict("DataTypeAltAz", {"alt": float, "az": float})


class MoveAltAzEvent(MoveEvent):
    """Event to be sent when moving to Alt/Az."""

    __module__ = "pyobs.events"

    def __init__(self, alt: float, az: float, **kwargs: Any):
        MoveEvent.__init__(self)
        self.data: DataTypeAltAz = {"alt": alt, "az": az}

    @property
    def alt(self) -> float:
        return self.data["alt"]

    @property
    def az(self) -> float:
        return self.data["az"]


__all__ = ["MoveEvent", "MoveRaDecEvent", "MoveAltAzEvent"]



from typing import Dict, Tuple, Any

from .base import SkyflatPriorities


class ConstSkyflatPriorities(SkyflatPriorities):
    """Constant flat priorities."""

    __module__ = "pyobs.utils.skyflats.priorities"

    def __init__(self, priorities: Dict[Tuple[str, Tuple[int, int]], float], *args: Any, **kwargs: Any):
        SkyflatPriorities.__init__(self)
        self._priorities = priorities

    async def __call__(self) -> Dict[Tuple[str, Tuple[int, int]], float]:
        return self._priorities


__all__ = ["ConstSkyflatPriorities"]



"""
    SleekXMPP: The Sleek XMPP Library
    Copyright (C) 2011 Nathanael C. Fritz, Dann Martens (TOMOTON).
    This file is part of SleekXMPP.

    See the file LICENSE for copying permission.
"""

from .RPC import MethodTimeout



from typing import Optional, Any
from typing_extensions import TypedDict

from .event import Event


DataType = TypedDict("DataType", {"current": str})


class FilterChangedEvent(Event):
    """Event to be sent when a filter has been changed."""

    __module__ = "pyobs.events"

    def __init__(self, current: str, **kwargs: Any):
        Event.__init__(self)
        self.data: DataType = {"current": current}

    @property
    def filter(self) -> str:
        return self.data["current"]


__all__ = ["FilterChangedEvent"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface


class IGain(Interface, metaclass=ABCMeta):
    """The camera supports setting of gain, to be used together with :class:`~pyobs.interfaces.ICamera`."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def set_gain(self, gain: float, **kwargs: Any) -> None:
        """Set the camera gain.

        Args:
            gain: New camera gain.

        Raises:
            ValueError: If gain could not be set.
        """
        ...

    @abstractmethod
    async def get_gain(self, **kwargs: Any) -> float:
        """Returns the camera binning.

        Returns:
            Current gain.
        """
        ...


__all__ = ["IGain"]



import logging
from typing import Any
import numpy as np
from astropy.coordinates import EarthLocation, AltAz

from pyobs.images import Image
from .applyoffsets import ApplyOffsets
from ..time import Time
from ...interfaces import ITelescope, IOffsetsAltAz

log = logging.getLogger(__name__)


class ApplyAltAzOffsets(ApplyOffsets):
    """Apply offsets from a given image to a given telescope."""

    __module__ = "pyobs.utils.offsets"

    def __init__(self, min_offset: float = 0.5, max_offset: float = 30, **kwargs: Any):
        """Initializes a new ApplyAltAzOffsets.

        Args:
            min_offset: Min offset in arcsec to move.
            max_offset: Max offset in arcsec to move.
        """
        ApplyOffsets.__init__(self, **kwargs)

        # store
        self._min_offset = min_offset
        self._max_offset = max_offset

    async def __call__(self, image: Image, telescope: ITelescope, location: EarthLocation) -> bool:
        """Take the pixel offsets stored in the meta data of the image and apply them to the given telescope.

        Args:
            image: Image to take offsets and WCS from.
            telescope: Telescope to apply offsets to.
            location: Observer location on Earth.

        Returns:
            Whether offsets have been applied successfully.
        """

        # telescope must be of type IAltAzOffsets
        tel = telescope
        if not isinstance(telescope, IOffsetsAltAz):
            log.error("Given telescope cannot handle Alt/Az offsets.")
            return False

        # get RA/Dec coordinates of center and center+offsets
        try:
            radec_center, radec_target = self._get_radec_center_target(image, location)
        except ValueError:
            log.warning("Could not get offsets from image meta.")
            return False

        # convert to Alt/Az
        frame = AltAz(obstime=Time(image.header["DATE-OBS"]), location=location)
        altaz_center = radec_center.transform_to(frame)
        altaz_target = radec_target.transform_to(frame)

        # get offset
        daz, dalt = altaz_center.spherical_offsets_to(altaz_target)
        log.info('Transformed to Alt/Az shift of dAlt=%.2f", dAz=%.2f".', dalt.arcsec, daz.arcsec)

        # get current offset
        cur_dalt, cur_daz = await telescope.get_offsets_altaz()

        # log it
        await self._log_offset(tel, "dalt", cur_dalt, dalt.degree, "daz", cur_daz, daz.degree)

        # too large or too small?
        diff = np.sqrt(dalt.arcsec ** 2.0 + daz.arcsec ** 2)
        if diff < self._min_offset:
            log.warning("Shift too small, skipping auto-guiding for now...")
            return False
        if diff > self._max_offset:
            log.warning("Shift too large, skipping auto-guiding for now...")
            return False

        # move offset
        log.info("Offsetting telescope...")
        await telescope.set_offsets_altaz(float(cur_dalt + dalt.degree), float(cur_daz + daz.degree))
        return True


__all__ = ["ApplyAltAzOffsets"]



import fnmatch
import glob
import logging
import os
import asyncio
from dataclasses import dataclass
from pathlib import PurePosixPath
from typing import Any, Optional, List, Tuple, AnyStr
from astropy.io import fits

from pyobs.modules import Module
from pyobs.utils.fits import format_filename

log = logging.getLogger(__name__)


@dataclass
class CurrentFile:
    filename: str
    data: AnyStr
    out_filename: Optional[str] = None
    hdu_list: Optional[fits.HDUList] = None


class ImageWatcher(Module):
    """Watch for new files and write them to all given destinations.

    Watches a path for new files and stores them in all given destinations. Only if all operations were successful,
    the file is deleted.
    """

    __module__ = "pyobs.modules.image"

    def __init__(
        self,
        watchpath: str,
        destinations: Optional[List[str]] = None,
        poll: bool = False,
        poll_interval: int = 5,
        wait_time: int = 10,
        pattern: str = "*",
        **kwargs: Any,
    ):
        """Create a new image watcher.

        Args:
            watchpath: Path to watch.
            destinations: Filename patterns for destinations.
            poll: If True, watchpath is polled instead of watched by inotify.
            poll_interval: Interval for polling in seconds, if poll is True.
            wait_time: Time in seconds between adding file to list and processing it.
        """
        Module.__init__(self, **kwargs)

        # add thread func
        self.add_background_task(self._worker)
        if poll:
            self.add_background_task(self._watch_poll)
        else:
            self.add_background_task(self._watch_inotify)

        # variables
        self._watchpath = watchpath
        self._notifier: Optional[Any] = None
        self._queue = asyncio.Queue[Tuple[str, asyncio.Future[None]]]()
        self._poll = poll
        self._poll_interval = poll_interval
        self._wait_time = wait_time
        self._pattern = pattern
        self.current_file: Optional[CurrentFile] = None

        # filename patterns
        if not destinations:
            raise ValueError("No filename patterns given for the destinations.")
        self._destinations = destinations

    async def _watch_inotify(self) -> None:
        from asyncinotify import Inotify, Mask  # type: ignore

        # get local directory
        local = await self.vfs.local_path(self._watchpath)

        # Context manager to close the inotify handle after use
        with Inotify() as inotify:
            # add watch on local directory
            inotify.add_watch(local, Mask.CLOSE_WRITE)

            # iterate events forever
            async for event in inotify:
                # get filename by replacing local with watchpath
                filename = str(event.path).replace(local, self._watchpath)

                # add file
                self.add_file(filename)

    async def _watch_poll(self) -> None:
        # init list
        files = set(await self.vfs.listdir(self._watchpath))

        # run forever
        path = PurePosixPath(self._watchpath)
        while True:
            # get new list
            new_files = await self.vfs.listdir(self._watchpath)

            # find all new files and add them
            for f in new_files:
                if f not in files:
                    print(str(path / f))
                    self.add_file(str(path / f))

            # store new list
            files = set(new_files)

    async def open(self) -> None:
        """Open image watcher."""
        await Module.open(self)

        # add all files from directory to queue
        for filename in await self.vfs.listdir(self._watchpath):
            self.add_file(os.path.join(self._watchpath, filename))

    async def close(self) -> None:
        """Close image watcher."""
        await Module.close(self)

        # stop watching
        if self._notifier:
            log.info("Stop watching directory...")
            self._notifier.stop()

    def add_file(self, filename: str) -> None:
        """Add a file to the file queue.

        Args:
            filename (str): Local filename of new file.
        """

        # check pattern
        if not fnmatch.fnmatch(filename, self._pattern):
            return

        # log and add file
        log.info("Adding new file %s...", filename)
        self._queue.put_nowait((filename, asyncio.create_task(asyncio.sleep(self._wait_time))))

    async def _worker(self) -> None:
        """Worker thread."""

        # run forever
        while True:
            # get next filename
            filename, future = await self._queue.get()

            # waiting for future, which is the wait time for new files
            await future
            log.info("Working on file %s...", filename)

            # better safe than sorry
            try:
                # get file data
                async with self.vfs.open_file(filename, "rb") as fd:
                    data = await fd.read()

                # try to load as fits file
                try:
                    fits_file = fits.HDUList.fromstring(data)
                except:
                    fits_file = None

                # fill current file
                self.current_file = CurrentFile(filename=filename, data=data, hdu_list=fits_file)

                # loop archive and upload
                success = True
                for pattern in self._destinations:
                    # if it contains {placeholders}, we assume it's a FITS file and format filename
                    if "{" in pattern and "}" in pattern and fits_file is not None:
                        # format filename
                        out_filename = format_filename(fits_file["SCI"].header, pattern)
                        if out_filename is None:
                            raise ValueError("Could not create name for file.")

                    else:
                        # no formatting, so just add filename to destination
                        out_filename = os.path.join(pattern, os.path.basename(filename))

                    # store it
                    log.info("Storing file as %s...", out_filename)
                    self.current_file.out_filename = out_filename
                    try:
                        async with self.vfs.open_file(out_filename, "wb") as fd:
                            await fd.write(data)
                    except:
                        log.exception("Error while copying file, skipping for now.")
                        success = False
                        break

                    # do extra processing
                    if not await self.process_extra(filename):
                        success = False
                        break

                # no success?
                if not success:
                    continue

                # close and delete files
                log.info("Removing file from watch directory...")
                if not await self.vfs.remove(filename):
                    log.warning("Could not delete %s.", filename)

                # cleanup extra
                await self.cleanup_extra(filename)

            except:
                log.exception("Something went wrong.")

    async def process_extra(self, filename: str) -> bool:
        """Can be overwritten by derived classes to do extra processing on files.
        All information are stored in self.current_file and can be checked against the given filename.

        Args:
            filename: Input name of original file.

        Returns:
            Whether processing was successful
        """
        return True

    async def cleanup_extra(self, filename: str) -> None:
        """Can be overwritten by derived classes to do clean up after successful copying.
        All information are stored in self.current_file and can be checked against the given filename.

        Args:
            filename: Input name of original file.
        """
        ...


__all__ = ["ImageWatcher"]



from abc import ABCMeta

from .IPointingAltAz import IPointingAltAz
from .IPointingRaDec import IPointingRaDec
from .IMotion import IMotion


class ITelescope(IMotion, IPointingAltAz, IPointingRaDec, metaclass=ABCMeta):
    """The module controls a telescope."""

    __module__ = "pyobs.interfaces"
    pass


__all__ = ["ITelescope"]



from typing import Optional, Dict, Any
from typing_extensions import TypedDict

from pyobs.events.event import Event
from pyobs.utils.enums import MotionStatus


DataType = TypedDict("DataType", {"status": str, "interfaces": Dict[str, str]})


class MotionStatusChangedEvent(Event):
    """Event to be sent when the motion status of a device has changed."""

    __module__ = "pyobs.events"

    def __init__(self, status: MotionStatus, interfaces: Optional[Dict[str, MotionStatus]] = None, **kwargs: Any):
        Event.__init__(self)
        self.data: DataType = {
            "status": status.value,
            "interfaces": {k: v.value for k, v in interfaces.items()} if interfaces is not None else {},
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Event:
        # get status
        if "status" not in d or not isinstance(d["status"], str):
            raise ValueError("Invalid type for status.")
        status: MotionStatus = MotionStatus(d["status"])

        # get interfaces
        interfaces: Optional[Dict[str, MotionStatus]] = {}
        if "interfaces" in d and isinstance(d["interfaces"], dict):
            interfaces = {k: MotionStatus(v) for k, v in d["interfaces"].items()}

        # return object
        return MotionStatusChangedEvent(status, interfaces)

    @property
    def status(self) -> MotionStatus:
        return MotionStatus(self.data["status"])

    @property
    def interfaces(self) -> Dict[str, MotionStatus]:
        return {k: MotionStatus(v) for k, v in self.data["interfaces"].items()}


__all__ = ["MotionStatusChangedEvent"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .IRunning import IRunning


class IStartStop(IRunning, metaclass=ABCMeta):
    """The module can be started and stopped."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def start(self, **kwargs: Any) -> None:
        """Starts a service."""
        ...

    @abstractmethod
    async def stop(self, **kwargs: Any) -> None:
        """Stops a service."""
        ...


__all__ = ["IStartStop"]



import fnmatch
from abc import ABCMeta, abstractmethod
from typing import Any, AnyStr, List


class VFSFile(metaclass=ABCMeta):
    """Base class for all VFS file classes."""

    __module__ = "pyobs.vfs"

    @abstractmethod
    async def close(self) -> None:
        ...

    @abstractmethod
    async def read(self, n: int = -1) -> AnyStr:
        ...

    @abstractmethod
    async def write(self, s: AnyStr) -> None:
        ...

    async def __aenter__(self) -> "VFSFile":
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.close()

    @staticmethod
    async def listdir(path: str, **kwargs: Any) -> List[str]:
        """Returns content of given path.

        Args:
            path: Path to list.
            kwargs: Parameters for specific file implementation (same as __init__).

        Returns:
            List of files in path.
        """
        raise NotImplementedError()

    @classmethod
    async def find(cls, path: str, pattern: str, **kwargs: Any) -> List[str]:
        """Find files by pattern matching.

        Args:
            path: Path to search in.
            pattern: Pattern to search for.

        Returns:
            List of found files.
        """

        # list files in dir
        files = await cls.listdir(path, **kwargs)

        # filter by pattern
        return list(filter(lambda f: fnmatch.fnmatch(f, pattern), files))

    @staticmethod
    async def remove(path: str, *args: Any, **kwargs: Any) -> bool:
        """Remove file at given path.

        Args:
            path: Path of file to delete.

        Returns:
            Success or not.
        """
        raise NotImplementedError()

    @classmethod
    async def exists(cls, path: str, *args: Any, **kwargs: Any) -> bool:
        """Checks, whether a given path or file exists.

        Args:
            path: Path to check.

        Returns:
            Whether it exists or not
        """
        raise NotImplementedError()


__all__ = ["VFSFile"]



"""
Modules for performing auto-guiding.
TODO: write doc
"""
__title__ = "Auto-guiding"

from ._baseguiding import BaseGuiding
from .autoguiding import AutoGuiding
from .scienceframeguiding import ScienceFrameAutoGuiding
from .acquisition import Acquisition
from .dummyacquisition import DummyAcquisition
from .dummyguiding import DummyAutoGuiding



from typing import List, Union, Any, Optional, Dict
import logging

from .publisher import Publisher


log = logging.getLogger(__name__)


class MultiPublisher(Publisher):
    """Forwards a message to multiple publishers."""

    def __init__(self, publishers: Optional[List[Union[Publisher, Dict[str, Any]]]] = None, **kwargs: Any):
        """Initialize new multi publisher.

        Args:
            publishers: Publishers to forwards messages to.
        """
        Publisher.__init__(self, **kwargs)

        # store
        self._publishers: List[Publisher] = [] if publishers is None else [self.add_child_object(p) for p in publishers]

    async def __call__(self, **kwargs: Any) -> None:
        """Publish the given results.

        Args:
            **kwargs: Results to publish.
        """

        # loop all publishers
        for p in self._publishers:
            # forward message
            await p(**kwargs)


__all__ = ["MultiPublisher"]



"""
The Comm object is responsible for all communication between modules (see :mod:`pyobs.modules`). The base
class for all implementations is :class:`~pyobs.comm.Comm`.

The user usually only has contact with the Comm object when writing the configuration for an existing module or when
developing a new module that needs to communicate with other modules.

In a configuration file, the Comm object is defined at top-level like this::

    comm:
        class: pyobs.comm.slixmpp.XmppComm

Except for a single parameter defined in :class:`~pyobs.comm.Comm`'s constructor, all parameters are defined in
derived classes.

The most convenient way for getting access to other modules' method is by using a :class:`~pyobs.comm.Proxy` object,
which can easily be obtained by using the :meth:`~pyobs.comm.Comm.proxy` method or the ``[]`` operator like this (if
the module named 'camera' implements the :class:`~pyobs.interfaces.ICamera` interface)::

    camera = comm['camera']
    camera.expose().wait()

Note that camera is now not of type :class:`~pyobs.interfaces.ICamera`.

Each :class:`~pyobs.modules.Module` that was configured with a Comm object (see :mod:`~pyobs.modules`) has
an attribute ``comm`` for easy access.

There is currently one one implementation of the Comm interface:

* :class:`~pyobs.comm.slixmpp.XmppComm` uses the XMPP protocol for communication.

.. seealso::

   Module :mod:`~pyobs.modules`
      Description for modules, to which Comm objects are usually assigned.
"""

from .comm import Comm
from .proxy import Proxy


__all__ = ["Comm", "Proxy"]



from sleekxmpp.xmlstream.stanzabase import ElementBase


class MethodTimeout(ElementBase):
    name = "methodTimeout"
    namespace = "jabber:iq:rpc"
    plugin_attrib = "method_timeout"
    interfaces = {"timeout"}
    subinterfaces = set(())
    plugin_attrib_map = {}
    plugin_tag_map = {}

    def get_timeout(self):
        return int(self._get_sub_text("timeout"))

    def set_timeout(self, timeout):
        return self._set_sub_text("timeout", str(timeout))



import io
import uuid
from typing import Optional, Any, AnyStr
from urllib.parse import urljoin
import logging
import aiohttp

from .file import VFSFile


log = logging.getLogger(__name__)


class HttpFile(VFSFile):
    """Wraps a file on a HTTP server that can be accessed via GET/POST.
    Especially useful in combination with :class:`~pyobs.modules.utils.HttpFileCache`."""

    __module__ = "pyobs.vfs"

    def __init__(
        self,
        name: str,
        mode: str = "r",
        download: Optional[str] = None,
        upload: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        verify_tls: bool = False,
        timeout: int = 30,
        **kwargs: Any,
    ):
        """Creates a new HTTP file.

        Args:
            name: Name of file.
            mode: Open mode (r/w).
            download: Base URL for downloading files. If None, no read access possible.
            upload: Base URL for uploading files. If None, no write access possible.
            username: Username for accessing the HTTP server.
            password: Password for accessing the HTTP server.
            verify_tls: Whether to verify TLS certificates.
            timeout: Timeout in seconds for uploading/downloading files.
        """

        # init
        io.RawIOBase.__init__(self)
        self._verify_tls = verify_tls
        self._timeout = timeout

        # auth
        self._auth = None
        if username is not None and password is not None:
            self._auth = aiohttp.BasicAuth(username, password)

        # filename is not allowed to start with a / or contain ..
        if name.startswith("/") or ".." in name:
            raise ValueError("Only files within root directory are allowed.")

        # build filename
        self.filename = name
        self.mode = mode
        self._buffer = b"" if "b" in self.mode else ""
        self._pos = 0
        self._open = True

        # URLs given?
        self._download_path = download
        self._upload_path = upload
        if "r" in self.mode and self._download_path is None:
            raise ValueError("No download URL given.")
        if "w" in self.mode and self._upload_path is None:
            raise ValueError("No upload URL given.")

    @property
    def url(self) -> str:
        """Returns URL of file."""
        if self._download_path is None:
            raise ValueError("No download URL given.")
        return urljoin(self._download_path, self.filename)

    async def _download(self) -> None:
        """For read access, download the file into a local buffer.

        Raises:
            FileNotFoundError: If file could not be found.
        """

        # do request
        async with aiohttp.ClientSession() as session:
            async with session.get(self.url, auth=self._auth, timeout=self._timeout) as response:
                # check response
                if response.status == 200:
                    # get data and return it
                    self._buffer = await response.read()
                elif response.status == 401:
                    log.error("Wrong credentials for downloading file.")
                    raise FileNotFoundError
                else:
                    log.error("Could not download file from filecache.")
                    raise FileNotFoundError

    async def read(self, n: int = -1) -> AnyStr:
        """Read number of bytes from stream.

        Args:
            n: Number of bytes to read. Read until end, if -1.

        Returns:
            Read bytes.
        """

        # load file
        if len(self._buffer) == 0 and "r" in self.mode:
            await self._download()

        # check size
        if n == -1:
            data = self._buffer
            self._pos = len(self._buffer) - 1
        else:
            # extract data to read
            data = self._buffer[self._pos : self._pos + n]
            self._pos += n

        # return data
        return data

    async def write(self, s: AnyStr) -> None:
        """Write data into the stream.

        Args:
            b: Bytes of data to write.
        """
        self._buffer += s

    async def close(self) -> None:
        """Close stream."""

        # write it?
        if "w" in self.mode and self._open:
            await self._upload()

        # set flag
        self._open = False

    async def _upload(self) -> None:
        """If in write mode, actually send the file to the HTTP server."""

        # filename given?
        filename = str(uuid.uuid4()) if self.filename is None else self.filename

        # check
        if self._upload_path is None:
            raise ValueError("No upload URL given.")

        # send data and return image ID
        async with aiohttp.ClientSession() as session:
            data = aiohttp.FormData()
            data.add_field("file", self._buffer, filename=self.filename)
            async with session.post(self._upload_path, auth=self._auth, data=data, timeout=self._timeout) as response:
                if response.status == 401:
                    log.error("Wrong credentials for uploading file.")
                    raise FileNotFoundError
                elif response.status != 200:
                    log.error("Could not upload file to filecache.")
                    raise FileNotFoundError


__all__ = ["HttpFile"]



from __future__ import annotations

import glob
import numpy as np
import astropy.units as u
from astropy.table import Table
from astropy.time import Time
from typing import Tuple, TYPE_CHECKING, Optional, Any, cast
from astropy.wcs import WCS
from astropy.io import fits
from numpy.typing import NDArray
from photutils.datasets import make_gaussian_sources_image
from photutils.datasets import make_noise_image
import logging

from pyobs.object import Object
from pyobs.utils.enums import ImageFormat
from pyobs.images import Image

if TYPE_CHECKING:
    from pyobs.utils.simulation import SimWorld


log = logging.getLogger(__name__)


class SimCamera(Object):
    """A simulated camera."""

    __module__ = "pyobs.utils.simulation"

    def __init__(
        self,
        world: "SimWorld",
        image_size: Optional[Tuple[int, int]] = None,
        pixel_size: float = 0.015,
        images: Optional[str] = None,
        max_mag: float = 20.0,
        seeing: float = 3.0,
        **kwargs: Any,
    ):
        """Inits a new camera.

        Args:
            world: World to use.
            image_size: Size of image.
            pixel_size: Square pixel size in mm.
            images: Filename pattern (e.g. /path/to/*.fits) for files to return instead of simulated images.
            max_mag: Maximum magnitude for sim.
            seeing: Seeing in arcsec.
        """
        Object.__init__(self, **kwargs)

        # store
        self.world = world
        self.telescope = world.telescope
        self.full_frame: Tuple[int, int, int, int] = (
            (0, 0, image_size[0], image_size[1]) if image_size is not None else (0, 0, 512, 512)
        )
        self.window = self.full_frame
        self.binning = (1, 1)
        self.pixel_size = pixel_size
        self.image_format = ImageFormat.INT16
        self.images = (
            [] if images is None else sorted(glob.glob(images)) if "*" in images or "?" in images else [images]
        )
        self._max_mag = max_mag
        self._seeing = seeing

        # private stuff
        self._catalog = None
        self._catalog_coords = None

    def get_image(self, exp_time: float, open_shutter: bool) -> Image:
        """Simulate an image.

        Args:
            exp_time: Exposure time in seconds.
            open_shutter: Whether the shutter is opened.

        Returns:
            numpy array with image.
        """

        # get now
        now = Time.now()

        # simulate or what?
        if self.images:
            # take image from list
            filename = self.images.pop(0)
            data = fits.getdata(filename)
            self.images.append(filename)

        else:
            # simulate
            data = self._simulate_image(exp_time, open_shutter)

        # create header
        hdr = self._create_header(exp_time, open_shutter, now, data)

        # return it
        return Image(data, header=hdr)

    def _simulate_image(self, exp_time: float, open_shutter: bool) -> NDArray[Any]:
        """Simulate an image.

        Args:
            exp_time: Exposure time in seconds.
            open_shutter: Whether the shutter is opened.

        Returns:
            numpy array with image.
        """

        # get shape for image
        shape = (int(self.window[3]), int(self.window[2]))

        # create image with Gaussian noise for BIAS
        data = make_noise_image(shape, distribution="gaussian", mean=10, stddev=1.0)

        # non-zero exposure time?
        if exp_time > 0:
            # add DARK
            data += make_noise_image(shape, distribution="gaussian", mean=exp_time / 1e4, stddev=exp_time / 1e5)

            # add stars and stuff
            if open_shutter:
                # get solar altitude
                sun_alt = self.world.sun_alt

                # get mean flatfield counts
                flat_counts = 30000 / np.exp(-1.28 * (4.209 + sun_alt)) * exp_time

                # create flat
                data += make_noise_image(shape, distribution="gaussian", mean=flat_counts, stddev=flat_counts / 10.0)

                # get catalog with sources
                sources = self._get_sources_table(exp_time)

                # filter out all sources outside FoV
                sources = sources[
                    (sources["x_mean"] > 0)
                    & (sources["x_mean"] < shape[1])
                    & (sources["y_mean"] > 0)
                    & (sources["y_mean"] < shape[0])
                ]

                # create image
                data += make_gaussian_sources_image(shape, sources)

        # saturate
        data[data > 65535] = 65535

        # finished
        return cast(NDArray[Any], data).astype(np.uint16)

    def _create_header(self, exp_time: float, open_shutter: float, time: Time, data: NDArray[Any]) -> fits.Header:
        # create header
        hdr = fits.Header()
        hdr["NAXIS1"] = data.shape[1]
        hdr["NAXIS2"] = data.shape[0]

        # set values
        hdr["DATE-OBS"] = (time.isot, "Date and time of start of exposure")
        hdr["EXPTIME"] = (exp_time, "Exposure time [s]")

        # binning
        hdr["XBINNING"] = hdr["DET-BIN1"] = (int(self.binning[0]), "Binning factor used on X axis")
        hdr["YBINNING"] = hdr["DET-BIN2"] = (int(self.binning[1]), "Binning factor used on Y axis")

        # window
        hdr["XORGSUBF"] = (int(self.window[0]), "Subframe origin on X axis")
        hdr["YORGSUBF"] = (int(self.window[1]), "Subframe origin on Y axis")

        # statistics
        hdr["DATAMIN"] = (float(np.min(data)), "Minimum data value")
        hdr["DATAMAX"] = (float(np.max(data)), "Maximum data value")
        hdr["DATAMEAN"] = (float(np.mean(data)), "Mean data value")

        # hardware
        hdr["TEL-FOCL"] = (self.telescope.focal_length, "Focal length [mm]")
        hdr["DET-PIXL"] = (self.pixel_size, "Size of detector pixels (square) [mm]")

        # finished
        return hdr

    def _get_catalog(self, fov: float) -> Table:
        """Returns GAIA catalog for current telescope coordinates."""
        # get catalog
        if self._catalog_coords is None or self._catalog_coords.separation(self.telescope.real_pos) > 10.0 * u.arcmin:
            from astroquery.utils.tap import TapPlus

            # get coordinates
            coords = self.telescope.real_pos

            # query TAP
            tap = TapPlus(url="https://gea.esac.esa.int/tap-server/tap")
            query = self._get_gaia_query(coords.ra.degree, coords.dec.degree, fov * 1.5)
            job = tap.launch_job(query)

            # get result table
            self._catalog = job.get_results()

        return self._catalog

    def _get_gaia_query(self, ra: float, dec: float, radius: float) -> str:
        # define query
        return f"""
                SELECT
                  TOP 1000
                  DISTANCE(
                    POINT('ICRS', ra, dec),
                    POINT('ICRS', {ra}, {dec})
                  ) as dist,
                  ra, dec, phot_g_mean_flux, phot_g_mean_mag
                FROM
                  gaiadr2.gaia_source
                WHERE
                  1 = CONTAINS(
                    POINT('ICRS', ra, dec),
                    CIRCLE('ICRS', {ra}, {dec}, {radius})
                  )
                  AND phot_g_mean_mag < {self._max_mag}
                ORDER BY
                  phot_g_mean_mag ASC
                """

    def _get_sources_table(self, exp_time: float) -> Table:
        """Create sources table."""

        # calculate cdelt1/2
        tmp = 360.0 / (2.0 * np.pi) * self.pixel_size / self.telescope.focal_length
        cdelt1, cdelt2 = tmp * self.binning[0], tmp * self.binning[1]
        log.info(
            "Plate scale is %.2f\"/px, image size is %.2f'x%.2f'.",
            cdelt1 * 3600,
            cdelt1 * 60 * self.window[2],
            cdelt2 * 60 * self.window[3],
        )

        # FoV
        fov = np.max(cdelt2 * np.array(self.full_frame[2:]))

        # get catalog
        cat = self._get_catalog(fov)

        # create WCS
        w = WCS(naxis=2)
        w.wcs.crpix = [self.window[3] / 2.0, self.window[2] / 2.0]
        w.wcs.cdelt = np.array([-cdelt1, cdelt2])
        w.wcs.crval = [self.telescope.real_pos.ra.degree, self.telescope.real_pos.dec.degree]
        w.wcs.ctype = ["RA---TAN", "DEC--TAN"]

        # set sigma to given seeing (in FWHM) in pixels
        fwhm = self._seeing / 3600.0 / cdelt1 / 2.3548

        # convert world to pixel coordinates
        cat["x"], cat["y"] = w.wcs_world2pix(cat["ra"], cat["dec"], 0)

        # get columns
        sources = cat["x", "y", "phot_g_mean_flux", "phot_g_mean_mag"]
        sources.rename_columns(["x", "y", "phot_g_mean_flux"], ["x_mean", "y_mean", "flux"])
        sources.add_column([fwhm] * len(sources), name="x_stddev")
        sources.add_column([fwhm] * len(sources), name="y_stddev")
        sources["flux"] *= exp_time

        """
        table['amplitude'] = [50, 70, 150, 210]
        table['x_mean'] = [160, 25, 150, 90]
        table['y_mean'] = [70, 40, 25, 60]
        table['x_stddev'] = [15.2, 5.1, 3., 8.1]
        table['y_stddev'] = [2.6, 2.5, 3., 4.7]
        table['theta'] = np.radians(np.array([145., 20., 0., 60.]))
        """

        # finished
        return sources


__all__ = ["SimCamera"]



"""
TODO: write doc
"""
__title__ = "FITS utilities"

import logging
import re
from typing import Union, Any, cast, Optional, List, Callable, Dict
import numpy as np
from astropy.io import fits
from numpy.typing import NDArray

from pyobs.utils.time import Time


log = logging.getLogger(__name__)


def fitssec(hdu: Any, keyword: str = "TRIMSEC") -> NDArray[Any]:
    """Trim an image to TRIMSEC or BIASSEC.

    Args:
        hdu: HDU to take data from.
        keyword: Header keyword for section.

    Returns:
        Numpy array with image data.
    """

    # keyword not given?
    if keyword not in hdu.header:
        # return whole data
        return cast(NDArray[Any], hdu.data)

    # get value of section
    sec = hdu.header[keyword]

    # split values
    s = sec[1:-1].split(",")
    x = s[0].split(":")
    y = s[1].split(":")
    x0 = int(x[0]) - 1
    x1 = int(x[1])
    y0 = int(y[0]) - 1
    y1 = int(y[1])

    # return data
    return cast(NDArray[Any], hdu.data[y0:y1, x0:x1])


class FilenameFormatter:
    def __init__(self, fmt: Union[str, List[str]], keys: Optional[Dict[str, Any]] = None):
        """Initializes a new filename formatter.

        Args:
            fmt: Filename format or list of formats. If list is given, first valid one is used.
            keys: Additional keys to pass to the formatter.
        """
        self.format = fmt
        self.keys: Dict[str, Any] = {} if keys is None else keys

        # define functions
        self.funcs: Dict[str, Callable[..., str]] = {
            "lower": self._format_lower,
            "time": self._format_time,
            "date": self._format_date,
            "filter": self._format_filter,
            "string": self._format_string,
            "type": self._format_type,
        }

    def _value(self, hdr: fits.Header, key: str) -> Any:
        """Returns value for given key.

        Args:
            hdr: fits.Header to take value from.
            key: Key to return value for.

        Returns:
            Value for given key.
        """

        # first check keys
        if key in self.keys:
            return self.keys[key]

        # then check header
        return hdr[key]

    def __call__(self, hdr: fits.Header) -> Optional[str]:
        """Formats a filename given a format template and a FITS header.

        Args:
            hdr: FITS header to take values from.

        Returns:
            Formatted filename.

        Raises:
            KeyError: If either keyword could not be found in header or method could not be found.
        """

        # make fmt a list
        if self.format is None:
            return None
        if isinstance(self.format, str):
            self.format = [self.format]

        # loop formats
        for f in self.format:
            try:
                # find all placeholders in format
                placeholders = re.findall("\{[\w\d_-]+(?:\|[\w\d_-]+\:?(?:[\w\d_-]+)*)?\}", f)
                if len(placeholders) == 0:
                    return f

                # create output
                output = f

                # loop all placeholders
                for ph in placeholders:
                    # call method and replace
                    output = output.replace(ph, str(self._format_placeholder(ph, hdr)))

                # finished
                return output

            except KeyError:
                # this format didn't work
                pass

        # still here?
        raise KeyError("No valid format found.")

    def _format_placeholder(self, placeholder: str, hdr: fits.Header) -> str:
        """Format a given placeholder.

        Args:
            placeholder: Placeholder to format.
            hdr: FITS header to take values from.

        Returns:
            Formatted placeholder.

        Raises:
            KeyError: If method or FITS header keyword don't exist.
        """

        # remove curly brackets
        key = placeholder[1:-1]
        method = None
        params = []

        # do we have a pipe in here?
        if "|" in key:
            key, method = key.split("|")

        # parameters for method?
        if method is not None and ":" in method:
            method, *params = method.split(":")

        # if no method is given, just replace
        if method is None:
            return str(self._value(hdr, key))

        else:
            # get function (may raise KeyError)
            func = self.funcs[method]

            # call method and replace
            return func(hdr, key, *params)

    def _format_lower(self, hdr: fits.Header, key: str) -> str:
        """Sets a given string to lowercase.

        Args:
            hdr: FITS header to take values from.
            key: The name of the FITS header key to use.

        Returns:
            Formatted string.
        """
        return str(self._value(hdr, key)).lower()

    def _format_time(self, hdr: fits.Header, key: str, delimiter: str = "-") -> str:
        """Formats time using the given delimiter.

        Args:
            hdr: FITS header to take values from.
            key: The name of the FITS header key to use.
            delimiter: Delimiter for time formatting.

        Returns:
            Formatted string.
        """
        fmt = "%H" + delimiter + "%M" + delimiter + "%S"
        date_obs = Time(self._value(hdr, key))
        return str(date_obs.datetime.strftime(fmt))

    def _format_date(self, hdr: fits.Header, key: str, delimiter: str = "-") -> str:
        """Formats date using the given delimiter.

        Args:
            hdr: FITS header to take values from.
            key: The name of the FITS header key to use.
            delimiter: Delimiter for date formatting.

        Returns:
            Formatted string.
        """
        fmt = "%Y" + delimiter + "%m" + delimiter + "%d"
        date_obs = Time(self._value(hdr, key))
        return str(date_obs.datetime.strftime(fmt))

    def _format_filter(self, hdr: fits.Header, key: str, image_type: str = "IMAGETYP", prefix: str = "-") -> str:
        """Formats a filter, prefixed by a given separator, only if the image type requires it.

        Args:
            hdr: FITS header to take values from.
            key: The name of the FITS header key to use.
            image_type: FITS header key for IMAGETYP.
            prefix: Prefix to add to filter.

        Returns:
            Formatted string.
        """
        it = hdr[image_type].lower()
        if it in ["light", "object", "flat", "skyflat"]:
            return str(prefix + self._value(hdr, key))
        else:
            return ""

    def _format_string(self, hdr: fits.Header, key: str, fmt: str) -> str:
        """Formats a string using Python string substitution.

        Args:
            hdr: FITS header to take values from.
            key: The name of the FITS header key to use.
            fmt: A Python string format like %d, %05d, or %4.1f.

        Returns:
            Formatted string.
        """
        fmt = "%" + fmt
        return str(fmt % self._value(hdr, key))

    def _format_type(self, hdr: fits.Header, key: str) -> str:
        """Formats an image type to a one-letter code.

        Args:
            hdr: FITS header to take values from.
            key: The name of the FITS header key to use.

        Returns:
            Formatted string.
        """
        if self._value(hdr, key) == "bias":
            return "b"
        elif self._value(hdr, key) == "skyflat":
            return "f"
        elif self._value(hdr, key) == "dark":
            return "d"
        else:
            return "e"


def format_filename(
    hdr: fits.Header, fmt: Union[str, List[str]], keys: Optional[Dict[str, Any]] = None
) -> Optional[str]:
    """Formats a filename given a format template and a FITS header.

    Args:
        hdr: FITS header to take values from.
        fmt: Filename format or list of formats. If multiple formats are given, the first valid one is used.
        keys: Additional keys to pass to the format string.

    Returns:
        Filename

    Raises:
        KeyError: If either keyword could not be found in header or method could not be found.
    """

    ff = FilenameFormatter(fmt, keys)
    return ff(hdr)


__all__ = ["format_filename", "FilenameFormatter", "fitssec"]



from __future__ import annotations
import asyncio
import logging
from typing import Tuple, List, Dict, Any, TYPE_CHECKING, Optional
from astropy.coordinates import SkyCoord
import astropy.units as u

from pyobs.events import FilterChangedEvent, OffsetsRaDecEvent
from pyobs.interfaces import IFocuser, IFitsHeaderBefore, IFilters, ITemperatures, IOffsetsRaDec
from pyobs.mixins.fitsnamespace import FitsNamespaceMixin
from pyobs.modules.telescope.basetelescope import BaseTelescope
from pyobs.modules import timeout
from pyobs.utils.enums import MotionStatus
from pyobs.utils.threads import LockWithAbort
from pyobs.utils.time import Time

if TYPE_CHECKING:
    from pyobs.utils.simulation import SimWorld


log = logging.getLogger(__name__)


class DummyTelescope(
    BaseTelescope, IOffsetsRaDec, IFocuser, IFilters, IFitsHeaderBefore, ITemperatures, FitsNamespaceMixin
):
    """A dummy telescope for testing."""

    __module__ = "pyobs.modules.telescope"

    def __init__(self, world: Optional[SimWorld] = None, **kwargs: Any):
        """Creates a new dummy telescope.

        Args:
            world: Optional SimWorld object.
        """
        BaseTelescope.__init__(self, **kwargs, motion_status_interfaces=["ITelescope", "IFocuser", "IFilters"])
        FitsNamespaceMixin.__init__(self, **kwargs)

        # init world and get telescope
        from pyobs.utils.simulation import SimWorld

        self._world = world if world is not None else self.add_child_object(SimWorld)
        self._telescope = self._world.telescope

        # automatically send status updates
        self._telescope.status_callback = self._change_motion_status

        # stuff
        self._lock_focus = asyncio.Lock()
        self._abort_focus = asyncio.Event()

    async def open(self) -> None:
        """Open module."""
        await BaseTelescope.open(self)

        # subscribe to events
        if self.comm:
            await self.comm.register_event(FilterChangedEvent)
            await self.comm.register_event(OffsetsRaDecEvent)

        # init status
        await self._change_motion_status(MotionStatus.IDLE)

    async def _move_radec(self, ra: float, dec: float, abort_event: asyncio.Event) -> None:
        """Actually starts tracking on given coordinates. Must be implemented by derived classes.

        Args:
            ra: RA in deg to track.
            dec: Dec in deg to track.
            abort_event: Event that gets triggered when movement should be aborted.

        Raises:
            MoveError: If telescope cannot be moved.
        """

        # start slewing
        await self.__move(ra, dec, abort_event)

    async def _move_altaz(self, alt: float, az: float, abort_event: asyncio.Event) -> None:
        """Actually moves to given coordinates. Must be implemented by derived classes.

        Args:
            alt: Alt in deg to move to.
            az: Az in deg to move to.
            abort_event: Event that gets triggered when movement should be aborted.

        Raises:
            MoveError: If telescope cannot be moved.
        """

        # alt/az coordinates to ra/dec
        coords = SkyCoord(
            alt=alt * u.degree, az=az * u.degree, obstime=Time.now(), location=self.location, frame="altaz"
        )
        icrs = coords.icrs

        # start slewing
        await self.__move(icrs.ra.degree, icrs.dec.degree, abort_event)

    async def __move(self, ra: float, dec: float, abort_event: asyncio.Event) -> None:
        """Simulate move.

        Args:
            ra: RA in deg to track.
            dec: Dec in deg to track.
            abort_event: Event that gets triggered when movement should be aborted.
        """

        # simulate slew
        self._telescope.move_ra_dec(SkyCoord(ra=ra * u.deg, dec=dec * u.deg, frame="icrs"))

        # wait for it
        while self._telescope.status == MotionStatus.SLEWING and not abort_event.is_set():
            await asyncio.sleep(1)

    async def get_focus(self, **kwargs: Any) -> float:
        """Return current focus.

        Returns:
            Current focus.
        """
        return self._telescope.focus

    @timeout(60)
    async def set_focus(self, focus: float, **kwargs: Any) -> None:
        """Sets new focus.

        Args:
            focus: New focus value.

        Raises:
            MoveError: If telescope cannot be moved.
        """

        # check
        if focus < 0 or focus > 100:
            raise ValueError("Invalid focus value.")

        # acquire lock
        async with LockWithAbort(self._lock_focus, self._abort_focus):
            log.info("Setting focus to %.2f..." % focus)
            await self._change_motion_status(MotionStatus.SLEWING, interface="IFocuser")
            ifoc = self._telescope.focus * 1.0
            dfoc = (focus - ifoc) / 300.0
            for i in range(300):
                # abort?
                if self._abort_focus.is_set():
                    raise InterruptedError("Setting focus was interrupted.")

                # move focus and sleep a little
                self._telescope.focus = ifoc + i * dfoc
                await asyncio.sleep(0.01)
            await self._change_motion_status(MotionStatus.POSITIONED, interface="IFocuser")
            self._telescope.focus = focus

    async def list_filters(self, **kwargs: Any) -> List[str]:
        """List available filters.

        Returns:
            List of available filters.
        """
        return self._telescope.filters

    async def get_filter(self, **kwargs: Any) -> str:
        """Get currently set filter.

        Returns:
            Name of currently set filter.
        """
        return self._telescope.filter_name

    async def set_filter(self, filter_name: str, **kwargs: Any) -> None:
        """Set the current filter.

        Args:
            filter_name: Name of filter to set.

        Raises:
            MoveError: If filter wheel cannot be moved.
        """

        # valid filter?
        if filter_name not in self._telescope.filters:
            raise ValueError("Invalid filter name.")

        # log and send event
        if filter_name != self._telescope.filter_name:
            # set it
            logging.info("Setting filter to %s", filter_name)
            await self._change_motion_status(MotionStatus.SLEWING, interface="IFilters")
            await asyncio.sleep(3)
            await self._change_motion_status(MotionStatus.POSITIONED, interface="IFilters")
            self._telescope.filter_name = filter_name

            # send event
            await self.comm.send_event(FilterChangedEvent(filter_name))
            logging.info("New filter set.")

    @timeout(60)
    async def init(self, **kwargs: Any) -> None:
        """Initialize telescope.

        Raises:
            InitError: If device could not be initialized.
        """

        # INIT, wait a little, then IDLE
        log.info("Initializing telescope...")
        await self._change_motion_status(MotionStatus.INITIALIZING)
        await asyncio.sleep(5)
        await self._change_motion_status(MotionStatus.IDLE)
        log.info("Telescope initialized.")

    @timeout(60)
    async def park(self, **kwargs: Any) -> None:
        """Park telescope.

        Raises:
            ParkError: If telescope could not be parked.
        """

        # PARK, wait a little, then PARKED
        log.info("Parking telescope...")
        await self._change_motion_status(MotionStatus.PARKING)
        await asyncio.sleep(5)
        await self._change_motion_status(MotionStatus.PARKED)
        log.info("Telescope parked.")

    async def set_offsets_radec(self, dra: float, ddec: float, **kwargs: Any) -> None:
        """Move an RA/Dec offset.

        Args:
            dra: RA offset in degrees.
            ddec: Dec offset in degrees.

        Raises:
            MoveError: If telescope cannot be moved.
        """

        log.info("Moving offset dra=%.5f, ddec=%.5f", dra, ddec)
        await self.comm.send_event(OffsetsRaDecEvent(ra=dra, dec=ddec))
        self._telescope.set_offsets(dra, ddec)

    async def get_offsets_radec(self, **kwargs: Any) -> Tuple[float, float]:
        """Get RA/Dec offset.

        Returns:
            Tuple with RA and Dec offsets.
        """
        return self._telescope.offsets

    async def get_radec(self, **kwargs: Any) -> Tuple[float, float]:
        """Returns current RA and Dec.

        Returns:
            Tuple of current RA and Dec in degrees.
        """
        return float(self._telescope.position.ra.degree), float(self._telescope.position.dec.degree)

    async def get_altaz(self, **kwargs: Any) -> Tuple[float, float]:
        """Returns current Alt and Az.

        Returns:
            Tuple of current Alt and Az in degrees.
        """
        if self.observer is not None:
            alt_az = self.observer.altaz(Time.now(), self._telescope.position)
            return float(alt_az.alt.degree), float(alt_az.az.degree)
        else:
            raise ValueError("No observer given.")

    async def get_fits_header_before(
        self, namespaces: Optional[List[str]] = None, **kwargs: Any
    ) -> Dict[str, Tuple[Any, str]]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """

        # fetch from BaseTelescope
        hdr = await BaseTelescope.get_fits_header_before(self)

        # focus
        hdr["TEL-FOCU"] = (self._telescope.focus, "Focus position [mm]")

        # finished
        return self._filter_fits_namespace(hdr, namespaces=namespaces, **kwargs)

    async def stop_motion(self, device: Optional[str] = None, **kwargs: Any) -> None:
        """Stop the motion.

        Args:
            device: Name of device to stop, or None for all.
        """
        pass

    async def get_focus_offset(self, **kwargs: Any) -> float:
        """Return current focus offset.

        Returns:
            Current focus offset.
        """
        return 0

    async def get_temperatures(self, **kwargs: Any) -> Dict[str, float]:
        """Returns all temperatures measured by this module.

        Returns:
            Dict containing temperatures.
        """

        return {"M1": 10.0, "M2": 12.0}

    async def set_focus_offset(self, offset: float, **kwargs: Any) -> None:
        log.error("Not implemented")

    async def is_ready(self, **kwargs: Any) -> bool:
        log.error("Not implemented")
        return True


__all__ = ["DummyTelescope"]



from __future__ import annotations

"""
TODO: Write docs
"""
__title__ = "Exceptions"

import asyncio
from collections import Coroutine
from typing import Optional, List, NamedTuple, Any, Tuple, Type, Dict, Callable
import time


class PyObsError(Exception):
    """Base class for all exceptions"""

    def __init__(self, message: Optional[str] = None):
        self.message = message

    def __str__(self) -> str:
        msg = f"<{self.__class__.__name__}>"
        if self.message is not None:
            msg += f" {self.message}"
        return msg


class _Meta(type):
    """Metaclass for defining exceptions."""

    def __call__(cls, *args: Any, **kwargs: Any) -> PyObsError:
        """Called when you call MyNewClass()"""
        exception: PyObsError = type.__call__(cls, *args, **kwargs)
        return handle_exception(exception)


#######################################


class ModuleError(PyObsError, metaclass=_Meta):
    pass


class MotionError(PyObsError, metaclass=_Meta):
    pass


class InitError(MotionError, metaclass=_Meta):
    pass


class ParkError(MotionError, metaclass=_Meta):
    pass


class MoveError(MotionError, metaclass=_Meta):
    pass


class GrabImageError(PyObsError, metaclass=_Meta):
    pass


#######################################


class RemoteError(PyObsError, metaclass=_Meta):
    """Exception for anything related to the communication between modules."""

    def __init__(self, module: str, message: Optional[str] = None):
        PyObsError.__init__(self, message)
        self.module = module


class RemoteTimeoutError(RemoteError, metaclass=_Meta):
    pass


class InvocationError(RemoteError, metaclass=_Meta):
    """Remote exception encapsulating basic exception from other module"""

    def __init__(self, module: str, exception: Exception):
        RemoteError.__init__(self, module, None)
        self.module = module
        a = ValueError()

        # never encapsulate a SevereError
        self.exception = exception.exception if isinstance(exception, SevereError) else exception

    def __str__(self) -> str:
        msg = f"<InvocationError> ({self.exception.__class__.__name__})"
        if hasattr(self.exception, "message"):
            if self.exception.message is not None:
                msg += f" {self.exception.message}"
        else:
            msg += f": {str(self.exception)}"
        return msg


#######################################


class SevereError(PyObsError):
    """Severe exception that is raised after multiple raised other exceptions."""

    def __init__(self, exception: PyObsError, module: Optional[str] = None):
        PyObsError.__init__(self, "A severe error has occurred.")
        self.module = module
        # never encapsulate a SevereError
        self.exception = exception.exception if isinstance(exception, SevereError) else exception


class LoggedException(NamedTuple):
    time: float
    exception: PyObsError


class ExceptionHandler(NamedTuple):
    exc_type: Type[PyObsError]
    limit: int
    timespan: Optional[float] = None
    module: Optional[str] = None
    callback: Optional[Callable[[PyObsError], Coroutine[Any, Any, None]]] = None
    throw: bool = False


#######################################


_local_exceptions: Dict[Type[PyObsError], List[LoggedException]] = {}
_remote_exceptions: Dict[Tuple[Type[PyObsError], str], List[LoggedException]] = {}
_handlers: List[ExceptionHandler] = []


def clear() -> None:
    _local_exceptions.clear()
    _remote_exceptions.clear()
    _handlers.clear()


def register_exception(
    exc_type: Type[PyObsError],
    limit: int,
    timespan: Optional[float] = None,
    module: Optional[str] = None,
    callback: Optional[Callable[[PyObsError], Coroutine[Any, Any, None]]] = None,
    throw: bool = False,
) -> None:
    _handlers.append(ExceptionHandler(exc_type, limit, timespan, module, callback, throw))


def handle_exception(exception: PyObsError) -> PyObsError:
    # get module and store exception
    module = exception.module if isinstance(exception, InvocationError) else None

    # store exception itself
    _store_exception(exception, module)

    # if there is a child exception, store it as well
    if hasattr(exception, "exception"):
        _store_exception(getattr(exception, "exception"), module)

    # now check, whether something is severe
    triggered_handlers = _check_severity()

    # filter triggered handlers by those that actually handle the exception
    handlers = list(filter(lambda h: isinstance(exception, h.exc_type), triggered_handlers))

    # check all handlers
    for h in handlers:
        # do we have a callback? then call it!
        if h.callback is not None:
            asyncio.create_task(h.callback(exception))

    # if we got any handlers triggered and throw is set on any, escalate to a SevereError
    if len(handlers) > 0 and any([h.throw for h in handlers]):
        return SevereError(exception=exception, module=module)

    # TODO: clean up old exceptions

    # else just return exception itself
    return exception


def _store_exception(exception: PyObsError, module: Optional[str]) -> None:
    # get all classes from mro
    for e in type(exception).__mro__:
        # only pyobs exceptions
        if not issubclass(e, PyObsError):
            continue

        # is it handled by any handler?
        if not any([e == h.exc_type for h in _handlers]):
            continue

        # log
        le = LoggedException(time=time.time(), exception=exception)

        # store it
        if module is None:
            # add to local exceptions
            if e not in _local_exceptions:
                _local_exceptions[e] = []
            _local_exceptions[e].append(le)

        else:
            # add to remote exceptions
            if (e, module) not in _remote_exceptions:
                _remote_exceptions[e, module] = []
            _remote_exceptions[e, module].append(le)


def _check_severity() -> List[ExceptionHandler]:
    """Checks all handlers against all raised exceptions and returns a list of triggered exception handlers.

    Returns:
        List of triggered handlers.
    """

    # loop all _handlers
    triggered: List[ExceptionHandler] = []
    for h in _handlers:
        # get all exceptions that this handler deals with
        exceptions = []
        if h.module is None:
            # add local exceptions
            if h.exc_type in _local_exceptions:
                exceptions.extend(_local_exceptions[h.exc_type])
        else:
            # add remote exceptions
            if (h.exc_type, h.module) in _remote_exceptions:
                exceptions.extend(_remote_exceptions[h.exc_type, h.module])

        # got a timespan?
        if h.timespan is None:
            # count all
            count = len(exceptions)

        else:
            # count all within timespan
            earliest = time.time() - h.timespan
            count = len(list(filter(lambda le: le.time >= earliest, exceptions)))

        # more than limit?
        if count >= h.limit:
            # add to list
            triggered.append(h)

    # return full list
    return triggered


#######################################



import asyncio
import copy
import json
import logging
import multiprocessing as mp
from typing import Union, List, Tuple, Any, Optional, Dict
import astroplan
from astroplan import ObservingBlock
from astropy.time import TimeDelta
import astropy.units as u

from pyobs.events.taskfinished import TaskFinishedEvent
from pyobs.events.taskstarted import TaskStartedEvent
from pyobs.events import GoodWeatherEvent, Event
from pyobs.utils.time import Time
from pyobs.interfaces import IStartStop, IRunnable
from pyobs.modules import Module
from pyobs.robotic import TaskArchive, TaskSchedule


log = logging.getLogger(__name__)


class Scheduler(Module, IStartStop, IRunnable):
    """Scheduler."""

    __module__ = "pyobs.modules.robotic"

    def __init__(
        self,
        tasks: Union[Dict[str, Any], TaskArchive],
        schedule: Union[Dict[str, Any], TaskSchedule],
        schedule_range: int = 24,
        safety_time: int = 60,
        twilight: str = "astronomical",
        trigger_on_task_started: bool = False,
        trigger_on_task_finished: bool = False,
        **kwargs: Any,
    ):
        """Initialize a new scheduler.

        Args:
            scheduler: Scheduler to use
            schedule_range: Number of hours to schedule into the future
            safety_time: If no ETA for next task to start exists (from current task, weather became good, etc), use
                         this time in seconds to make sure that we don't schedule for a time when the scheduler is
                         still running
            twilight: astronomical or nautical
            trigger_on_task_started: Whether to trigger a re-calculation of schedule, when task has started.
            trigger_on_task_finishes: Whether to trigger a re-calculation of schedule, when task has finished.
        """
        Module.__init__(self, **kwargs)

        # get scheduler
        self._task_archive = self.add_child_object(tasks, TaskArchive)
        self._schedule = self.add_child_object(schedule, TaskSchedule)

        # store
        self._schedule_range = schedule_range
        self._safety_time = safety_time
        self._twilight = twilight
        self._running = True
        self._initial_update_done = False
        self._need_update = False
        self._trigger_on_task_started = trigger_on_task_started
        self._trigger_on_task_finished = trigger_on_task_finished

        # time to start next schedule from
        self._schedule_start: Optional[Time] = None

        # ID of currently running task, and current (or last if finished) block
        self._current_task_id = None
        self._last_task_id = None

        # blocks
        self._blocks: List[ObservingBlock] = []

        # update thread
        self.add_background_task(self._schedule_worker)
        self.add_background_task(self._update_worker)

    async def open(self) -> None:
        """Open module."""
        await Module.open(self)

        # subscribe to events
        if self.comm:
            await self.comm.register_event(TaskStartedEvent, self._on_task_started)
            await self.comm.register_event(TaskFinishedEvent, self._on_task_finished)
            await self.comm.register_event(GoodWeatherEvent, self._on_good_weather)

    async def start(self, **kwargs: Any) -> None:
        """Start scheduler."""
        self._running = True

    async def stop(self, **kwargs: Any) -> None:
        """Stop scheduler."""
        self._running = False

    async def is_running(self, **kwargs: Any) -> bool:
        """Whether scheduler is running."""
        return self._running

    async def _update_worker(self) -> None:
        # time of last change in blocks
        last_change = None

        # run forever
        while True:
            # not running?
            if self._running is False:
                await asyncio.sleep(1)
                continue

            # got new time of last change?
            t = await self._task_archive.last_changed()
            if last_change is None or last_change < t:
                # get schedulable blocks and sort them
                log.info("Found update in schedulable block, downloading them...")
                blocks = sorted(
                    await self._task_archive.get_schedulable_blocks(),
                    key=lambda x: json.dumps(x.configuration, sort_keys=True),
                )
                log.info("Downloaded %d schedulable block(s).", len(blocks))

                # compare new and old lists
                removed, added = self._compare_block_lists(self._blocks, blocks)

                # schedule update
                self._need_update = True

                # no changes?
                if len(removed) == 0 and len(added) == 0:
                    # no need to re-schedule
                    log.info("No change in list of blocks detected.")
                    self._need_update = False

                # has only the current block been removed?
                log.info("Removed: %d, added: %d", len(removed), len(added))
                if len(removed) == 1:
                    log.info(
                        "Found 1 removed block with ID %d. Last task ID was %s, current is %s.",
                        removed[0].target.name,
                        str(self._last_task_id),
                        str(self._current_task_id),
                    )
                if len(removed) == 1 and len(added) == 0 and removed[0].target.name == self._last_task_id:
                    # no need to re-schedule
                    log.info("Only one removed block detected, which is the one currently running.")
                    self._need_update = False

                # store blocks
                self._blocks = blocks

                # schedule update
                if self._need_update:
                    log.info("Triggering scheduler run...")

                # remember now
                last_change = Time.now()
                self._initial_update_done = True

            # sleep a little
            await asyncio.sleep(5)

    @staticmethod
    def _compare_block_lists(
        blocks1: List[ObservingBlock], blocks2: List[ObservingBlock]
    ) -> Tuple[List[ObservingBlock], List[ObservingBlock]]:
        """Compares two lists of ObservingBlocks and returns two lists, containing those that are missing in list 1
        and list 2, respectively.

        Args:
            blocks1: First list of blocks.
            blocks2: Second list of blocks.

        Returns:
            (tuple): Tuple containing:
                unique1:  Blocks that exist in blocks1, but not in blocks2.
                unique2:  Blocks that exist in blocks2, but not in blocks1.
        """

        # get dictionaries with block names
        names1 = {b.target.name: b for b in blocks1}
        names2 = {b.target.name: b for b in blocks2}

        # find elements in names1 that are missing in names2 and vice versa
        additional1 = set(names1.keys()).difference(names2.keys())
        additional2 = set(names2.keys()).difference(names1.keys())

        # get blocks for names and return them
        unique1 = [names1[n] for n in additional1]
        unique2 = [names2[n] for n in additional2]
        return unique1, unique2

    async def _schedule_worker(self) -> None:
        # run forever
        while True:
            # need update?
            if self._need_update and self._initial_update_done:
                # reset need for update
                self._need_update = False

                try:
                    # prepare scheduler
                    blocks, start, end, constraints = await self._prepare_schedule()

                    # schedule
                    scheduled_blocks = await self._schedule_blocks(blocks, start, end, constraints)

                    # finish schedule
                    await self._finish_schedule(scheduled_blocks, start)

                except ValueError as e:
                    log.warning(str(e))

            # sleep a little
            await asyncio.sleep(1)

    async def _prepare_schedule(self) -> Tuple[List[ObservingBlock], Time, Time, List[Any]]:
        """TaskSchedule blocks."""

        # only global constraint is the night
        if self._twilight == "astronomical":
            constraints = [astroplan.AtNightConstraint.twilight_astronomical()]
        elif self._twilight == "nautical":
            constraints = [astroplan.AtNightConstraint.twilight_nautical()]
        else:
            raise ValueError("Unknown twilight type.")

        # make shallow copies of all blocks and loop them
        copied_blocks = [copy.copy(block) for block in self._blocks]
        for block in copied_blocks:
            # astroplan's PriorityScheduler expects lower priorities to be more important, so calculate
            # 1000 - priority
            block.priority = 1000.0 - block.priority
            if block.priority < 0:
                block.priority = 0

            # it also doesn't match the requested observing windows exactly, so we make them a little smaller.
            for constraint in block.constraints:
                if isinstance(constraint, astroplan.TimeConstraint):
                    constraint.min += 30 * u.second
                    constraint.max -= 30 * u.second

        # get start time for scheduler
        start = self._schedule_start
        now_plus_safety = Time.now() + self._safety_time * u.second
        if start is None or start < now_plus_safety:
            # if no ETA exists or is in the past, use safety time
            start = now_plus_safety

        # get running scheduled block, if any
        if self._current_task_id is None:
            log.info("No running block found.")
            running_task = None
        else:
            # get running task from archive
            log.info("Trying to find running block in current schedule...")
            now = Time.now()
            tasks = await self._schedule.get_schedule(now, now, include_running=True)
            if self._current_task_id in tasks:
                running_task = tasks[self._current_task_id]
            else:
                log.info("Running block not found in last schedule.")
                running_task = None

        # if start is before end time of currently running block, change that
        if running_task is not None:
            log.info("Found running block that ends at %s.", running_task.end)

            # get block end plus some safety
            block_end = running_task.end + 10.0 * u.second
            if start < block_end:
                start = block_end
                log.info("Start time would be within currently running block, shifting to %s.", start.isot)

        # calculate end time
        end = start + TimeDelta(self._schedule_range * u.hour)

        # remove currently running block and filter by start time
        blocks: List[ObservingBlock] = []
        for b in filter(lambda x: x.configuration["request"]["id"] != self._current_task_id, copied_blocks):
            time_constraint_found = False
            # loop all constraints
            for c in b.constraints:
                if isinstance(c, astroplan.TimeConstraint):
                    # we found a time constraint
                    time_constraint_found = True

                    # does the window start before the end of the scheduling range?
                    if c.min < end:
                        # yes, store block and break loop
                        blocks.append(b)
                        break
            else:
                # loop has finished without breaking
                # if no time constraint has been found, we still take the block
                if time_constraint_found is False:
                    blocks.append(b)

        # if need new update, skip here
        if self._need_update:
            raise ValueError("Not running scheduler, since update was requested.")

        # no blocks found?
        if len(blocks) == 0:
            await self._schedule.set_schedule([], start)
            raise ValueError("No blocks left for scheduling.")

        # return all
        return blocks, start, end, constraints

    async def _schedule_blocks(
        self, blocks: List[ObservingBlock], start: Time, end: Time, constraints: List[Any]
    ) -> List[ObservingBlock]:

        # run actual scheduler in separate process and wait for it
        qout: mp.Queue = mp.Queue()
        p = mp.Process(target=self._schedule_process, args=(blocks, start, end, constraints, qout))
        p.start()

        # wait for process to finish
        # note that the process only finishes, when the queue is empty! so we have to poll the queue first
        # and then the process.
        loop = asyncio.get_running_loop()
        scheduled_blocks: List[ObservingBlock] = await loop.run_in_executor(None, qout.get, True)
        await loop.run_in_executor(None, p.join)
        return scheduled_blocks

    async def _finish_schedule(self, scheduled_blocks: List[ObservingBlock], start: Time) -> None:
        # if need new update, skip here
        if self._need_update:
            log.info("Not using scheduler results, since update was requested.")
            return

        # update
        await self._schedule.set_schedule(scheduled_blocks, start)
        if len(scheduled_blocks) > 0:
            log.info("Finished calculating schedule for %d block(s):", len(scheduled_blocks))
            for i, block in enumerate(scheduled_blocks, 1):
                log.info(
                    "  #%d: %s to %s (%.1f)",
                    block.configuration["request"]["id"],
                    block.start_time.strftime("%H:%M:%S"),
                    block.end_time.strftime("%H:%M:%S"),
                    block.priority,
                )
        else:
            log.info("Finished calculating schedule for 0 blocks.")

    def _schedule_process(
        self,
        blocks: List[ObservingBlock],
        start: Time,
        end: Time,
        constraints: List[Any],
        scheduled_blocks: mp.Queue,
    ) -> None:
        """Actually do the scheduling, usually run in a separate process."""

        # log it
        log.info("Calculating schedule for %d schedulable block(s) starting at %s...", len(blocks), start)

        # we don't need any transitions
        transitioner = astroplan.Transitioner()

        # create scheduler
        scheduler = astroplan.PriorityScheduler(constraints, self.observer, transitioner=transitioner)

        # run scheduler
        time_range = astroplan.Schedule(start, end)
        schedule = scheduler(blocks, time_range)

        # put scheduled blocks in queue
        scheduled_blocks.put(schedule.scheduled_blocks)

    async def run(self, **kwargs: Any) -> None:
        """Trigger a re-schedule."""
        self._need_update = True

    async def _on_task_started(self, event: Event, sender: str) -> bool:
        """Re-schedule when task has started and we can predict its end.

        Args:
            event: The task started event.
            sender: Who sent it.
        """
        if not isinstance(event, TaskStartedEvent):
            return False

        # store it
        self._current_task_id = event.id
        self._last_task_id = event.id

        # trigger?
        if self._trigger_on_task_started:
            # get ETA in minutes
            eta = (event.eta - Time.now()).sec / 60
            log.info("Received task started event with ETA of %.0f minutes, triggering new scheduler run...", eta)

            # set it
            self._need_update = True
            self._schedule_start = event.eta

        return True

    async def _on_task_finished(self, event: Event, sender: str) -> bool:
        """Reset current task, when it has finished.

        Args:
            event: The task finished event.
            sender: Who sent it.
        """
        if not isinstance(event, TaskFinishedEvent):
            return False

        # reset current task
        self._current_task_id = None

        # trigger?
        if self._trigger_on_task_finished:
            # get ETA in minutes
            log.info("Received task finished event, triggering new scheduler run...")

            # set it
            self._need_update = True
            self._schedule_start = Time.now()

        return True

    async def _on_good_weather(self, event: Event, sender: str) -> bool:
        """Re-schedule on incoming good weather event.

        Args:
            event: The good weather event.
            sender: Who sent it.
        """
        if not isinstance(event, GoodWeatherEvent):
            return False

        # get ETA in minutes
        eta = (event.eta - Time.now()).sec / 60
        log.info("Received good weather event with ETA of %.0f minutes, triggering new scheduler run...", eta)

        # set it
        self._need_update = True
        self._schedule_start = event.eta
        return True

    async def abort(self, **kwargs: Any) -> None:
        pass


__all__ = ["Scheduler"]



import asyncio
from abc import ABCMeta, abstractmethod
from typing import Dict, Any, Tuple, Union, List, Optional
from astropy.coordinates import SkyCoord, ICRS, AltAz
import astropy.units as u
import logging

from pyobs.events import MoveRaDecEvent, MoveAltAzEvent
from pyobs.interfaces import ITelescope, IFitsHeaderBefore
from pyobs.modules import Module
from pyobs.mixins import MotionStatusMixin, WeatherAwareMixin, WaitForMotionMixin
from pyobs.modules import timeout
from pyobs.utils.enums import MotionStatus
from pyobs.utils.threads import LockWithAbort
from pyobs.utils.time import Time
from pyobs.utils import exceptions as exc

log = logging.getLogger(__name__)


class BaseTelescope(
    WeatherAwareMixin, MotionStatusMixin, WaitForMotionMixin, ITelescope, IFitsHeaderBefore, Module, metaclass=ABCMeta
):
    """Base class for telescopes."""

    __module__ = "pyobs.modules.telescope"

    def __init__(
        self,
        fits_headers: Optional[Dict[str, Any]] = None,
        min_altitude: float = 10,
        wait_for_dome: Optional[str] = None,
        **kwargs: Any,
    ):
        """Initialize a new base telescope.

        Args:
            fits_headers: Additional FITS headers to send.
            min_altitude: Minimal altitude for telescope.
            wait_for_dome: Name of dome module to wait for.
        """
        Module.__init__(self, **kwargs)

        # store
        self._fits_headers = fits_headers if fits_headers is not None else {}
        self._min_altitude = min_altitude

        # some multi-threading stuff
        self._lock_moving = asyncio.Lock()
        self._abort_move = asyncio.Event()

        # celestial status
        self._celestial_headers: Dict[str, Any] = {}

        # add thread func
        self.add_background_task(self._celestial, True)

        # init mixins
        WeatherAwareMixin.__init__(self, **kwargs)
        MotionStatusMixin.__init__(self, **kwargs)
        WaitForMotionMixin.__init__(
            self,
            wait_for_modules=None if wait_for_dome is None else [wait_for_dome],
            wait_for_timeout=60000,
            wait_for_states=[MotionStatus.POSITIONED, MotionStatus.TRACKING],
        )

        # register exception
        exc.register_exception(exc.MotionError, 3, timespan=600, callback=self._default_remote_error_callback)

    async def open(self) -> None:
        """Open module."""
        await Module.open(self)

        # open mixins
        await WeatherAwareMixin.open(self)
        await MotionStatusMixin.open(self)

    @abstractmethod
    async def _move_radec(self, ra: float, dec: float, abort_event: asyncio.Event) -> None:
        """Actually starts tracking on given coordinates. Must be implemented by derived classes.

        Args:
            ra: RA in deg to track.
            dec: Dec in deg to track.
            abort_event: Event that gets triggered when movement should be aborted.

        Raises:
            MoveError: If telescope cannot be moved.
        """
        ...

    @timeout(1200)
    async def move_radec(self, ra: float, dec: float, **kwargs: Any) -> None:
        """Starts tracking on given coordinates.

        Args:
            ra: RA in deg to track.
            dec: Dec in deg to track.

        Raises:
            MoveError: If telescope cannot be moved.
        """

        # do nothing, if initializing, parking or parked
        if await self.get_motion_status() in [MotionStatus.INITIALIZING, MotionStatus.PARKING, MotionStatus.PARKED]:
            return

        # check observer
        if self.observer is None:
            raise ValueError("No observer given.")

        # to alt/az
        ra_dec = SkyCoord(ra * u.deg, dec * u.deg, frame=ICRS)
        alt_az = self.observer.altaz(Time.now(), ra_dec)

        # check altitude
        if alt_az.alt.degree < self._min_altitude:
            raise ValueError("Destination altitude below limit.")

        # acquire lock
        async with LockWithAbort(self._lock_moving, self._abort_move):
            # log and event
            await self._change_motion_status(MotionStatus.SLEWING)
            log.info(
                "Moving telescope to RA=%s (%.5f°), Dec=%s (%.5f°)...",
                ra_dec.ra.to_string(sep=":", unit=u.hour, pad=True),
                ra,
                ra_dec.dec.to_string(sep=":", unit=u.deg, pad=True),
                dec,
            )
            await self.comm.send_event(MoveRaDecEvent(ra=ra, dec=dec))

            # track telescope
            await self._move_radec(ra, dec, abort_event=self._abort_move)
            log.info("Reached destination")

            # move dome, if exists
            await self._wait_for_motion(self._abort_move)

            # finish slewing
            await self._change_motion_status(MotionStatus.TRACKING)

            # update headers now
            asyncio.create_task(self._update_celestial_headers())
            log.info("Finished moving telescope.")

    @abstractmethod
    async def _move_altaz(self, alt: float, az: float, abort_event: asyncio.Event) -> None:
        """Actually moves to given coordinates. Must be implemented by derived classes.

        Args:
            alt: Alt in deg to move to.
            az: Az in deg to move to.
            abort_event: Event that gets triggered when movement should be aborted.

        Raises:
            MoveError: If telescope cannot be moved.
        """
        ...

    @timeout(1200)
    async def move_altaz(self, alt: float, az: float, **kwargs: Any) -> None:
        """Moves to given coordinates.

        Args:
            alt: Alt in deg to move to.
            az: Az in deg to move to.

        Raises:
            MoveError: If telescope cannot be moved.
        """

        # do nothing, if initializing, parking or parked
        if await self.get_motion_status() in [MotionStatus.INITIALIZING, MotionStatus.PARKING, MotionStatus.PARKED]:
            return

        # check altitude
        if alt < self._min_altitude:
            raise ValueError("Destination altitude below limit.")

        # acquire lock
        async with LockWithAbort(self._lock_moving, self._abort_move):
            # log and event
            log.info("Moving telescope to Alt=%.2f°, Az=%.2f°...", alt, az)
            await self.comm.send_event(MoveAltAzEvent(alt=alt, az=az))
            await self._change_motion_status(MotionStatus.SLEWING)

            # move telescope
            await self._move_altaz(alt, az, abort_event=self._abort_move)
            log.info("Reached destination")

            # move dome, if exists
            await self._wait_for_motion(self._abort_move)

            # finish slewing
            await self._change_motion_status(MotionStatus.POSITIONED)

            # update headers now
            asyncio.create_task(self._update_celestial_headers())
            log.info("Finished moving telescope.")

    async def get_fits_header_before(
        self, namespaces: Optional[List[str]] = None, **kwargs: Any
    ) -> Dict[str, Tuple[Any, str]]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """

        # define base header
        hdr: Dict[str, Union[Any, Tuple[Any, str]]] = {}

        # positions
        try:
            ra, dec = await self.get_radec()
            coords_ra_dec = SkyCoord(ra=ra * u.deg, dec=dec * u.deg, frame=ICRS)
            alt, az = await self.get_altaz()
            coords_alt_az = SkyCoord(alt=alt * u.deg, az=az * u.deg, frame=AltAz)

        except Exception as e:
            log.warning("Could not fetch telescope position: %s", e)
            coords_ra_dec, coords_alt_az = None, None

        # set coordinate headers
        if coords_ra_dec is not None:
            hdr["TEL-RA"] = (float(coords_ra_dec.ra.degree), "Right ascension of telescope [degrees]")
            hdr["TEL-DEC"] = (float(coords_ra_dec.dec.degree), "Declination of telescope [degrees]")
        if coords_alt_az is not None:
            hdr["TEL-ALT"] = (float(coords_alt_az.alt.degree), "Telescope altitude [degrees]")
            hdr["TEL-AZ"] = (float(coords_alt_az.az.degree), "Telescope azimuth [degrees]")
            hdr["TEL-ZD"] = (90.0 - hdr["TEL-ALT"][0], "Telescope zenith distance [degrees]")
            hdr["AIRMASS"] = (float(coords_alt_az.secz.value), "Airmass of observation start")

        # convert to sexagesimal
        if coords_ra_dec is not None:
            hdr["RA"] = (str(coords_ra_dec.ra.to_string(sep=":", unit=u.hour, pad=True)), "Right ascension of object")
            hdr["DEC"] = (str(coords_ra_dec.dec.to_string(sep=":", unit=u.deg, pad=True)), "Declination of object")

        # site location
        if self.observer is not None:
            hdr["LATITUDE"] = (float(self.observer.location.lat.degree), "Latitude of telescope [deg N]")
            hdr["LONGITUD"] = (float(self.observer.location.lon.degree), "Longitude of telescope [deg E]")
            hdr["HEIGHT"] = (float(self.observer.location.height.value), "Altitude of telescope [m]")

        # add static fits headers
        for key, value in self._fits_headers.items():
            hdr[key] = tuple(value)

        # add celestial headers
        for key, value in self._celestial_headers.items():
            hdr[key] = tuple(value)

        # finish
        return hdr

    async def _celestial(self) -> None:
        """Thread for continuously calculating positions and distances to celestial objects like moon and sun."""

        # wait a little
        await asyncio.sleep(10)

        # run until closing
        while True:
            # update headers
            await self._update_celestial_headers()

            # sleep a little
            await asyncio.sleep(30)

    async def _update_celestial_headers(self) -> None:
        """Calculate positions and distances to celestial objects like moon and sun."""
        # get now
        now = Time.now()
        alt: Optional[float]
        az: Optional[float]

        # no observer?
        if self.observer is None:
            return

        # get telescope alt/az
        try:
            alt, az = await self.get_altaz()
            tel_altaz = SkyCoord(alt=alt * u.deg, az=az * u.deg, frame="altaz")
        except:
            alt, az, tel_altaz = None, None, None

        # get current moon and sun information
        moon_altaz = self.observer.moon_altaz(now)
        moon_frac = self.observer.moon_illumination(now)
        sun_altaz = self.observer.sun_altaz(now)

        # calculate distance to telescope
        moon_dist = tel_altaz.separation(moon_altaz) if tel_altaz is not None else None
        sun_dist = tel_altaz.separation(sun_altaz) if tel_altaz is not None else None

        # store it
        self._celestial_headers = {
            "MOONALT": (float(moon_altaz.alt.degree), "Lunar altitude"),
            "MOONFRAC": (float(moon_frac), "Fraction of the moon illuminated"),
            "MOONDIST": (None if moon_dist is None else float(moon_dist.degree), "Lunar distance from target"),
            "SUNALT": (float(sun_altaz.alt.degree), "Solar altitude"),
            "SUNDIST": (None if sun_dist is None else float(sun_dist.degree), "Solar Distance from Target"),
        }


__all__ = ["BaseTelescope"]



import asyncio
from typing import Any
import logging
import numpy as np

from .photometry import Photometry
from pyobs.images import Image

log = logging.getLogger(__name__)


class SepPhotometry(Photometry):
    """Perform photometry using SEP."""

    __module__ = "pyobs.images.processors.photometry"

    def __init__(
        self,
        threshold: float = 1.5,
        minarea: int = 5,
        deblend_nthresh: int = 32,
        deblend_cont: float = 0.005,
        clean: bool = True,
        clean_param: float = 1.0,
        **kwargs: Any,
    ):
        """Initializes a wrapper for SEP. See its documentation for details.

        Highly inspired by LCO's wrapper for SEP, see:
        https://github.com/LCOGT/banzai/blob/master/banzai/photometry.py

        Args:
            threshold: Threshold pixel value for detection.
            minarea: Minimum number of pixels required for detection.
            deblend_nthresh: Number of thresholds used for object deblending.
            deblend_cont: Minimum contrast ratio used for object deblending.
            clean: Perform cleaning?
            clean_param: Cleaning parameter (see SExtractor manual).
            *args:
            **kwargs:
        """
        Photometry.__init__(self, **kwargs)

        # store
        self.threshold = threshold
        self.minarea = minarea
        self.deblend_nthresh = deblend_nthresh
        self.deblend_cont = deblend_cont
        self.clean = clean
        self.clean_param = clean_param

    async def __call__(self, image: Image) -> Image:
        """Do aperture photometry on given image.

        Args:
            image: Image to do aperture photometry on.

        Returns:
            Image with attached catalog.
        """
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, SepPhotometry._photometry, image)

    @staticmethod
    def _photometry(image: Image) -> Image:
        import sep
        from pyobs.images.processors.detection import SepSourceDetection

        # check data
        if image.data is None:
            log.warning("No data found in image.")
            return image
        if image.catalog is None:
            log.warning("No catalog found in image.")
            return image

        # no mask?
        mask = image.mask if image.mask is not None else np.ones(image.data.shape, dtype=bool)

        # remove background
        data, bkg = SepSourceDetection.remove_background(image.data, mask)

        # fetch catalog
        sources = image.catalog.copy()

        # match SEP conventions
        x, y = sources["x"] - 1, sources["y"] - 1

        # get gain
        gain = image.header["DET-GAIN"] if "DET-GAIN" in image.header else None

        # perform aperture photometry for diameters of 1" to 8"
        for diameter in [1, 2, 3, 4, 5, 6, 7, 8]:
            if image.pixel_scale is not None:
                flux, fluxerr, flag = sep.sum_circle(
                    data, x, y, diameter / 2.0 / image.pixel_scale, mask=image.mask, err=image.uncertainty, gain=gain
                )
                sources["fluxaper{0}".format(diameter)] = flux
                sources["fluxerr{0}".format(diameter)] = fluxerr

            else:
                sources["fluxaper{0}".format(diameter)] = 0
                sources["fluxerr{0}".format(diameter)] = 0

        # average background at each source
        # since SEP sums up whole pixels, we need to do the same on an image of ones for the background_area
        bkgflux, fluxerr, flag = sep.sum_ellipse(
            bkg.back(), x, y, sources["a"], sources["b"], np.pi / 2.0, 2.5 * sources["kronrad"], subpix=1
        )
        background_area, _, _ = sep.sum_ellipse(
            np.ones(shape=bkg.back().shape),
            x,
            y,
            sources["a"],
            sources["b"],
            np.pi / 2.0,
            2.5 * sources["kronrad"],
            subpix=1,
        )
        sources["background"] = bkgflux
        sources["background"][background_area > 0] /= background_area[background_area > 0]

        # pick columns for catalog
        new_columns = [
            "fluxaper1",
            "fluxerr1",
            "fluxaper2",
            "fluxerr2",
            "fluxaper3",
            "fluxerr3",
            "fluxaper4",
            "fluxerr4",
            "fluxaper5",
            "fluxerr5",
            "fluxaper6",
            "fluxerr6",
            "fluxaper7",
            "fluxerr7",
            "fluxaper8",
            "fluxerr8",
            "background",
        ]
        cat = sources[image.catalog.colnames + new_columns]

        # copy image, set catalog and return it
        img = image.copy()
        img.catalog = cat
        return img


__all__ = ["SepPhotometry"]



import asyncio
import os
from typing import Optional, Any, AnyStr, List
import paramiko
import paramiko.sftp

from .file import VFSFile


class SSHFile(VFSFile):
    """VFS wrapper for a file that can be accessed over a SFTP connection."""

    __module__ = "pyobs.vfs"

    def __init__(
        self,
        name: str,
        mode: str = "r",
        bufsize: int = -1,
        hostname: Optional[str] = None,
        port: int = 22,
        username: Optional[str] = None,
        password: Optional[str] = None,
        keyfile: Optional[str] = None,
        root: Optional[str] = None,
        mkdir: bool = True,
        **kwargs: Any,
    ):
        """Open/create a file over a SSH connection.

        Args:
            name: Name of file.
            mode: Open mode.
            bufsize: Size of buffer size for SFTP connection.
            hostname: Name of host to connect to.
            port: Port on host to connect to.
            username: Username to log in on host.
            password: Password for username.
            keyfile: Path to SSH key on local machine.
            root: Root directory on host.
            mkdir: Whether or not to automatically create directories.
        """

        # no root given?
        if root is None:
            raise ValueError("No root directory given.")

        # filename is not allowed to start with a / or contain ..
        if name.startswith("/") or ".." in name:
            raise ValueError("Only files within root directory are allowed.")

        # build filename
        self.filename = name
        full_path = os.path.join(root, name)

        # check
        if hostname is None:
            raise ValueError("No hostname given.")

        # connect
        self._ssh = paramiko.SSHClient()
        self._ssh.load_system_host_keys()
        self._ssh.connect(hostname, port=port, username=username, password=password, key_filename=keyfile)
        self._sftp = self._ssh.open_sftp()

        # need to create directory?
        path = os.path.dirname(full_path)
        try:
            self._sftp.chdir(path)
        except IOError:
            if mkdir:
                self._sftp.mkdir(path)
            else:
                raise ValueError("Cannot write into sub-directory with disabled mkdir option.")

        # open file
        self._fd = self._sftp.file(full_path, mode)

    async def close(self) -> None:
        """Close file."""
        self._sftp.close()
        self._ssh.close()

    async def read(self, n: int = -1) -> AnyStr:
        return self._fd.read(n)

    async def write(self, s: AnyStr) -> None:
        self._fd.write(s)

    @staticmethod
    async def listdir(path: str, **kwargs: Any) -> List[str]:
        """Returns content of given path.

        Args:
            path: Path to list.
            kwargs: Parameters for specific file implementation (same as __init__).

        Returns:
            List of files in path.
        """

        # connect
        ssh = paramiko.SSHClient()
        ssh.load_system_host_keys()
        ssh.connect(
            kwargs["hostname"],
            port=kwargs["port"] if "port" in kwargs else 22,
            username=kwargs["username"],
            password=kwargs["password"] if "password" in kwargs else None,
            key_filename=kwargs["keyfile"] if "keyfile" in kwargs else None,
        )
        sftp = ssh.open_sftp()

        # list files in path
        loop = asyncio.get_running_loop()
        files = await loop.run_in_executor(None, sftp.listdir, os.path.join(kwargs["root"], path))

        # disconnect and return list
        sftp.close()
        ssh.close()
        return files


__all__ = ["SSHFile"]



import asyncio
from functools import partial
from pathlib import PureWindowsPath
from typing import Optional, Any, AnyStr, List
import logging

from .file import VFSFile


# set logging for smbprotocol package
logging.getLogger("smbprotocol").setLevel(logging.WARNING)


class SMBFile(VFSFile):
    """VFS wrapper for a file that can be accessed over a SMB connection.

    Requires smbprotocol package to work.
    """

    __module__ = "pyobs.vfs"

    def __init__(
        self,
        name: str,
        mode: str = "r",
        hostname: Optional[str] = None,
        share: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        root: Optional[str] = None,
        mkdir: bool = True,
        **kwargs: Any,
    ):
        """Open/create a file over a SSH connection.

        Args:
            name: Name of file.
            mode: Open mode.
            hostname: Name of host to connect to.
            share: Share to access on server.
            username: Username to log in on host.
            password: Password for username.
            keyfile: Path to SSH key on local machine.
            root: Root directory on host.
            mkdir: Whether or not to automatically create directories.
        """
        try:
            import smbclient  # type: ignore
        except ModuleNotFoundError:
            raise ValueError("Module smbprotocol not found, please install package.")

        # no root given?
        if root is None:
            raise ValueError("No root directory given.")

        # filename is not allowed to start with a / or contain ..
        if name.startswith("/") or ".." in name:
            raise ValueError("Only files within root directory are allowed.")

        # check
        if hostname is None or share is None:
            raise ValueError("No hostname/share given.")

        # store connection details
        self._hostname = hostname
        self._share = share
        self._username = username
        self._password = password

        # build filename
        self.filename = name
        self._full_path = PureWindowsPath(rf"\\{hostname}\{share}\\") / root / name

        # need to create directory?
        path = str(self._full_path.parent)
        try:
            smbclient.lstat(path)
        except IOError:
            if mkdir:
                smbclient.mkdir(path)
            else:
                raise ValueError("Cannot write into sub-directory with disabled mkdir option.")

        # open file
        self._fd = smbclient.open_file(
            str(self._full_path), mode=mode, username=self._username, password=self._password
        )

    async def close(self) -> None:
        """Close file."""
        self._fd.close()

    async def read(self, n: int = -1) -> AnyStr:
        return self._fd.read(n)

    async def write(self, s: AnyStr) -> None:
        self._fd.write(s)

    @staticmethod
    async def listdir(path: str, **kwargs: Any) -> List[str]:
        """Returns content of given path.

        Args:
            path: Path to list.
            kwargs: Parameters for specific file implementation (same as __init__).

        Returns:
            List of files in path.
        """
        import smbclient

        # get settings
        hostname = kwargs["hostname"]
        share = kwargs["share"]
        username = kwargs["username"]
        password = kwargs["password"]
        root = kwargs["root"]

        # get path and return list
        network = PureWindowsPath(rf"\\{hostname}\{share}\\") / root / path
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(
            None, partial(smbclient.listdir, str(network), username=username, password=password)
        )


__all__ = ["SMBFile"]



from __future__ import annotations
import contextlib
import time
import asyncio
import inspect
from asyncio import Task
from collections import Coroutine
from typing import TypeVar, Optional, List, Any, cast, Union

from pyobs.utils.types import cast_response_to_real


async def event_wait(evt: asyncio.Event, timeout: float = 1.0) -> bool:
    # suppress TimeoutError because we'll return False in case of timeout
    with contextlib.suppress(asyncio.TimeoutError):
        await asyncio.wait_for(evt.wait(), timeout)
    return evt.is_set()


async def acquire_lock(lock: asyncio.Lock, timeout: float = 1.0) -> bool:
    # suppress TimeoutError because we'll return False in case of timeout
    try:
        await asyncio.wait_for(lock.acquire(), timeout)
        return True
    except asyncio.TimeoutError:
        return False


class Future(asyncio.Future):
    def __init__(self, empty: bool = False, signature: Optional[inspect.Signature] = None, *args: Any, **kwargs: Any):
        asyncio.Future.__init__(self, *args, **kwargs)

        """Init new base future."""
        self.timeout: Optional[float] = None
        self.signature: Optional[inspect.Signature] = signature

        # already set?
        if empty:
            # fire event
            self.set_result(None)

    def set_timeout(self, timeout: float) -> None:
        """
        Sets a new timeout for the method call.
        """
        self.timeout = timeout

    def get_timeout(self) -> Optional[float]:
        """
        Returns async timeout.
        """
        return self.timeout

    def _wait_for_time(self, timeout: float = 0) -> None:
        """Waits a little.

        Args:
            time: Time to wait in seconds.
        """
        start = time.time()
        while not self.done() or time.time() - start > timeout:
            return
        raise TimeoutError

    def __await__(self) -> Any:
        # not finished? need to wait.
        if not self.done():
            try:
                # wait some 10s first
                self._wait_for_time(10)

            except TimeoutError:
                # got an additional timeout?
                if self.timeout is not None and self.timeout > 10:
                    # we already waited 10s, so subtract it
                    self._wait_for_time(self.timeout - 10.0)

        # not done? yield!
        if not self.done():
            self._asyncio_future_blocking = True
            yield self  # This tells Task to wait for completion.

        # still not done? raise exception.
        if not self.done():
            raise RuntimeError("await wasn't used with future")

        # get result
        result = self.result()

        # all ok, return value
        if self.signature is not None:
            # cast response to real types
            return cast_response_to_real(result, self.signature)
        else:
            return result

    @staticmethod
    async def wait_all(futures: List[Optional[Union[Future, Coroutine[Any, Any, Any], Task[Any]]]]) -> List[Any]:
        return [await fut for fut in futures if fut is not None]


__all__ = ["Future", "event_wait", "acquire_lock"]



import asyncio
import logging
from typing import Any, Optional

from pyobs.modules.pointing._baseguiding import BaseGuiding
from pyobs.images.meta.exptime import ExpTime
from pyobs.images.processors.detection import SepSourceDetection
from pyobs.interfaces import IExposureTime, IImageType, IImageGrabber
from pyobs.utils.enums import ImageType
from pyobs.utils.parallel import event_wait

log = logging.getLogger(__name__)


class AutoGuiding(BaseGuiding):
    """An auto-guiding system."""

    __module__ = "pyobs.modules.guiding"

    def __init__(self, exposure_time: float = 1.0, **kwargs: Any):
        """Initializes a new auto guiding system.

        Args:
            exposure_time: Initial exposure time in seconds.
        """
        BaseGuiding.__init__(self, **kwargs)

        # store
        self._default_exposure_time = exposure_time
        self._exposure_time: Optional[float] = None
        self._source_detection = SepSourceDetection()

        # add thread func
        self.add_background_task(self._auto_guiding)

    async def set_exposure_time(self, exposure_time: float, **kwargs: Any) -> None:
        """Set the exposure time for the auto-guider.

        Args:
            exposure_time: Exposure time in secs.
        """
        log.info("Setting exposure time to %ds...", exposure_time)
        self._default_exposure_time = exposure_time
        self._exposure_time = None
        self._loop_closed = False
        await self._reset_guiding(enabled=self._enabled)

    async def _auto_guiding(self) -> None:
        # exposure time
        self._exposure_time = self._default_exposure_time

        # run until closed
        while True:
            # not running?
            if not self._enabled:
                await asyncio.sleep(1)
                continue

            try:
                # get camera
                camera = await self.proxy(self._camera, IImageGrabber)

                # take image
                if isinstance(camera, IExposureTime):
                    # set exposure time
                    log.info("Taking image with an exposure time of %.2fs...", self._exposure_time)
                    await camera.set_exposure_time(self._exposure_time)
                else:
                    log.info("Taking image...")
                if isinstance(camera, IImageType):
                    await camera.set_image_type(ImageType.OBJECT)
                filename = await camera.grab_image(broadcast=False)

                # download image
                image = await self.vfs.read_image(filename)

                # process it
                log.info("Processing image...")
                processed_image = await self._process_image(image)
                log.info("Done.")

                # new exposure time?
                if processed_image is not None and processed_image.has_meta(ExpTime):
                    self._exposure_time = processed_image.get_meta(ExpTime).exptime

                # sleep a little
                await asyncio.sleep(self._min_interval)

            except Exception as e:
                log.error("An error occurred: ", e)
                await asyncio.sleep(5)


__all__ = ["AutoGuiding"]



from abc import ABCMeta, abstractmethod
from typing import Tuple, Any

from .interface import Interface


class ILatLon(Interface, metaclass=ABCMeta):
    """The module can move to general Lat/Lon coordinates, which have to be async defined by the module itself.
    Usually combined with :class:`~pyobs.interfaces.ITelescope`."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def move_latlon(self, lat: float, lon: float, **kwargs: Any) -> None:
        """Moves to given coordinates.

        Args:
            lat: Latitude in deg to move to.
            lon: Longitude in deg to move to.

        Raises:
            ValueError: If device could not move.
        """
        ...

    @abstractmethod
    async def get_latlon(self, **kwargs: Any) -> Tuple[float, float]:
        """Returns current Latitude and Longitude.

        Returns:
            Tuple of current Latitude and Longitude in degrees.
        """
        ...


__all__ = ["ILatLon"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .IAbortable import IAbortable


class IRunnable(IAbortable, metaclass=ABCMeta):
    """The module has some action that can be started remotely."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def run(self, **kwargs: Any) -> None:
        """Perform module task"""
        ...


__all__ = ["IRunnable"]



from typing import Union, Dict, Tuple, List, Any
from astropy.time import TimeDelta
import astropy.units as u

from pyobs.object import get_object
from pyobs.utils.archive import Archive
from pyobs.utils.time import Time
from .base import SkyflatPriorities
from ...enums import ImageType


class ArchiveSkyflatPriorities(SkyflatPriorities):
    """Calculate flat priorities from an archive."""

    __module__ = "pyobs.utils.skyflats.priorities"

    def __init__(
        self,
        archive: Union[Dict[str, Any], Archive],
        site: str,
        instrument: str,
        filter_names: List[str],
        binnings: List[int],
        *args,
        **kwargs,
    ):
        SkyflatPriorities.__init__(self)
        self._archive = get_object(archive, Archive)
        self._filter_names = filter_names
        self._binnings = binnings
        self._site = site
        self._instrument = instrument

    async def __call__(self) -> Dict[Tuple[str, Tuple[int, int]], float]:
        # get all reduced skyflat frames of the last 100 days
        now = Time.now()
        frames = await self._archive.list_frames(
            start=now - TimeDelta(100 * u.day),
            end=now,
            site=self._site,
            instrument=self._instrument,
            image_type=ImageType.SKYFLAT,
            rlevel=1,
        )

        # get priorities
        from_archive: Dict[Tuple[str, int], float] = {}
        for f in frames:
            # get number of days since flat was taken, which is our priority
            prio = (now - f.dateobs).sec / 86400.0

            # get key in priorities
            key = (f.filter_name, f.binning)

            # need to update it?
            if key not in from_archive or prio < from_archive[key]:
                from_archive[key] = prio

        # create priorities
        priorities: Dict[Tuple[str, Tuple[int, int]], float] = {}
        for fn in self._filter_names:
            for b in self._binnings:
                priorities[fn, (b, b)] = from_archive[fn, b] if (fn, b) in from_archive else 100.0

        # finished
        return priorities


__all__ = ["ArchiveSkyflatPriorities"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .IAbortable import IAbortable
from .IImageGrabber import IImageGrabber
from pyobs.utils.enums import ExposureStatus


class ICamera(IAbortable, IImageGrabber, metaclass=ABCMeta):
    """The module controls a camera."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def get_exposure_status(self, **kwargs: Any) -> ExposureStatus:
        """Returns the current status of the camera, which is one of 'idle', 'exposing', or 'readout'.

        Returns:
            Current status of camera.
        """
        ...

    @abstractmethod
    async def abort(self, **kwargs: Any) -> None:
        """Aborts the current exposure and sequence.

        Raises:
            ValueError: If exposure could not be aborted.
        """
        ...

    @abstractmethod
    async def get_exposure_progress(self, **kwargs: Any) -> float:
        """Returns the progress of the current exposure in percent.

        Returns:
            Progress of the current exposure in percent.
        """
        ...


__all__ = ["ICamera"]
